from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional, Union

from valor.enums import AnnotationType, MetricType, ROUGEType, TaskType
from valor.schemas.filters import Filter


@dataclass
class EvaluationParameters:
    """
    Defines parameters for evaluation methods.

    Attributes
    ----------
    task_type: TaskType
        The task type of a given evaluation.
    label_map: Optional[List[List[List[str]]]]
        Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
    metrics_to_return: List[MetricType], optional
        The list of metrics to compute, store, and return to the user.
    llm_api_params: Dict[str, str | dict], optional
        A dictionary of parameters for the LLM API.
    convert_annotations_to_type: AnnotationType | None = None
        The type to convert all annotations to.
    iou_thresholds_to_compute: List[float], optional
        A list of floats describing which Intersection over Unions (IoUs) to use when calculating metrics (i.e., mAP).
    iou_thresholds_to_return: List[float], optional
        A list of floats describing which Intersection over Union (IoUs) thresholds to calculate a metric for. Must be a subset of `iou_thresholds_to_compute`.
    recall_score_threshold: float, default=0
        The confidence score threshold for use when determining whether to count a prediction as a true positive or not while calculating Average Recall.
    pr_curve_iou_threshold: float, optional
        The IOU threshold to use when calculating precision-recall curves for object detection tasks. Defaults to 0.5.
    pr_curve_max_examples: int
        The maximum number of datum examples to store when calculating PR curves.
    bleu_weights: list[float], optional
        The weights to use when calculating BLEU scores.
    rouge_types: list[ROUGEType]
        A list of rouge types to calculate. Options are ['rouge1', 'rouge2', 'rougeL', 'rougeLsum'], where `rouge1` is unigram-based scoring, `rouge2` is bigram-based scoring, `rougeL` is scoring based on sentences (i.e., splitting on "." and ignoring "\n"), and `rougeLsum` is scoring based on splitting the text using "\n".
    rouge_use_stemmer: bool
        If True, uses Porter stemmer to strip word suffixes.
    """

    task_type: TaskType
    label_map: Optional[List[List[List[str]]]] = None
    metrics_to_return: Optional[List[MetricType]] = None
    llm_api_params: Optional[Dict[str, Union[str, dict]]] = None

    convert_annotations_to_type: Optional[AnnotationType] = None
    iou_thresholds_to_compute: Optional[List[float]] = None
    iou_thresholds_to_return: Optional[List[float]] = None
    recall_score_threshold: float = 0
    pr_curve_iou_threshold: float = 0.5
    pr_curve_max_examples: int = 1
    bleu_weights: Optional[List[float]] = None
    rouge_types: Optional[List[ROUGEType]] = None
    rouge_use_stemmer: Optional[bool] = None


@dataclass
class EvaluationRequest:
    """
    An evaluation request.

    Defines important attributes of the API's `EvaluationRequest`.

    Attributes
    ----------
    dataset_names : List[str]
        The list of datasets we want to evaluate by name.
    model_names : List[str]
        The list of models we want to evaluate by name.
    filters : dict
        The filter object used to define what the model(s) is evaluating against.
    parameters : EvaluationParameters
        Any parameters that are used to modify an evaluation method.
    """

    dataset_names: Union[str, List[str]]
    model_names: Union[str, List[str]]
    parameters: EvaluationParameters
    filters: Filter = field(default_factory=Filter)

    def __post_init__(self):
        if isinstance(self.filters, dict):
            self.filters = Filter(**self.filters)
        elif self.filters is None:
            self.filters = Filter()

        if isinstance(self.parameters, dict):
            self.parameters = EvaluationParameters(**self.parameters)

    def to_dict(self) -> dict:
        """
        Converts the request into a JSON-compatible dictionary.
        """
        return {
            "dataset_names": self.dataset_names,
            "model_names": self.model_names,
            "parameters": asdict(self.parameters),
            "filters": self.filters.to_dict(),
        }



import datetime
from typing import Optional
from unittest.mock import Mock

import pytest

from valor import Evaluation, enums, schemas

try:
    import pandas as pd  # noqa: F401
except ModuleNotFoundError:
    pd = None


@pytest.mark.skipif(pd is None, reason="pandas package is not installed")
def test_to_dataframe():
    def _generate_metric(
        type: str,
        parameters: Optional[dict] = None,
        value: Optional[float] = None,
        label: Optional[dict] = None,
    ):
        return dict(type=type, parameters=parameters, value=value, label=label)

    df = Evaluation(
        id=1,
        dataset_names=["dataset1"],
        model_name="model1",
        filters=schemas.Filter(),
        parameters=schemas.EvaluationParameters(
            task_type=enums.TaskType.CLASSIFICATION,
        ),
        status=enums.EvaluationStatus.DONE,
        metrics=[
            _generate_metric(
                "d",
                parameters={"x": 0.123, "y": 0.987},
                value=0.3,
                label={"key": "k1", "value": "v2"},
            ),
            _generate_metric("a", value=0.99),
            _generate_metric("b", value=0.3),
            _generate_metric(
                "c", parameters={"x": 0.123, "y": 0.987}, value=0.3
            ),
            _generate_metric(
                "d",
                parameters={"x": 0.123, "y": 0.987},
                value=0.3,
                label={"key": "k1", "value": "v1"},
            ),
        ],
        confusion_matrices=[],
        created_at=datetime.datetime(
            year=2024, month=1, day=1, microsecond=1
        ).strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
        connection=Mock(),
        meta={},
    ).to_dataframe()

    df_str = """                                        value
    evaluation                              1
    type parameters               label
    a    "n/a"                    n/a        0.99
    b    "n/a"                    n/a        0.30
    c    {"x": 0.123, "y": 0.987} n/a        0.30
    d    {"x": 0.123, "y": 0.987} k1: v1     0.30
                                  k1: v2     0.30"""

    assert str(df).replace(" ", "") == df_str.replace(" ", "")



import json

from requests import Response, exceptions


class ClientException(Exception):
    def __init__(self, resp: Response):
        self.status_code = resp.status_code
        self.detail = resp.json()["detail"]
        super().__init__(str(self.detail))


class ClientAlreadyConnectedError(Exception):
    def __init__(self):
        super().__init__("Client already connected.")


class ClientNotConnectedError(Exception):
    def __init__(self):
        super().__init__("Client not connected.")


class ClientConnectionFailed(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class ServiceUnavailable(ClientException):
    """
    Raises an exception if the Valor service is unavailble.
    """

    pass


class DatasetAlreadyExistsError(ClientException):
    """
    Raises an exception if the user tries to create a dataset with a name that already exists.
    """

    pass


class DatasetDoesNotExistError(ClientException):
    """
    Raises an exception if the user tries to manipulate a dataset that doesn't exist.
    """

    pass


class DatasetFinalizedError(ClientException):
    """
    Raises an exception if the user tries to add groundtruths to a dataset that has already been finalized.
    """

    pass


class DatasetNotFinalizedError(ClientException):
    """
    Raises an exception if the user tries to process a dataset that hasn't been finalized.
    """

    pass


class ModelAlreadyExistsError(ClientException):
    """
    Raises an exception if the user tries to create a model using a name that already exists in the database.
    """

    pass


class ModelDoesNotExistError(ClientException):
    """
    Raises an exception if the user tries to manipulate a model that doesn't exist.
    """

    pass


class ModelFinalizedError(ClientException):
    """
    Raises an exception if the user tries to add predictions to a model that has been finalized.
    """

    pass


class ModelNotFinalizedError(ClientException):
    """
    Raises an exception if the user tries to manipulate a model that hasn't been finalized.
    """

    pass


class DatumDoesNotExistError(ClientException):
    """
    Raises an exception if the user tries to manipulate a datum that doesn't exist.
    """

    pass


class DatumAlreadyExistsError(ClientException):
    """
    Raises an exception if the user tries to create a datum that already exists.

    """

    pass


class DatumsAlreadyExistError(ClientException):
    """
    Raises an exception if the user tries to create a datum that already exists.

    """

    pass


class AnnotationAlreadyExistsError(ClientException):
    """
    Raises an exception if the user tries to create a annotation for a datum that already has annotation(s).
    """

    pass


class PredictionDoesNotExistError(ClientException):
    """
    Raises an exception if a prediction does not exist for a given model, dataset, and datum
    """

    pass


class PredictionAlreadyExistsError(ClientException):
    """
    Raises an exception if a prediction already exists for a given model, dataset, and datum
    """

    pass


class EvaluationRequestError(ClientException):
    """
    Raises an exception if an evaluation request fails validation.
    """

    def __init__(self, resp: Response):
        self.status_code = resp.status_code
        details = json.loads(json.loads(resp.json()["detail"])["detail"])
        self.detail = details["description"]
        self.errors = details["errors"]
        error_types = {error["name"] for error in self.errors}
        Exception.__init__(self, str(self.detail), str(error_types))


class EvaluationDoesNotExist(Exception):
    """
    Raises an exception if an evaluation does not exist.
    """

    def __init__(self, evaluation_id: int):
        super().__init__(f"Evaluation '{evaluation_id}' does not exist.")


def raise_client_exception(resp: Response):
    try:
        resp_json = resp.json()
        try:
            error_dict = json.loads(resp_json["detail"])
            cls_name = error_dict["name"]
            if cls_name in globals() and issubclass(
                globals()[cls_name], ClientException
            ):
                raise globals()[cls_name](resp)
            else:
                raise ClientException(resp)
        except (TypeError, json.JSONDecodeError):
            raise ClientException(resp)
    except (exceptions.JSONDecodeError, KeyError):
        resp.raise_for_status()



import numpy as np
import pytest

from valor.schemas import Box, MultiPolygon, Point, Polygon, Raster


def test_point():
    # valid
    p1 = Point((1, 1))
    p2 = Point((1.0, 1.0))
    p3 = Point((1.0, 0.99))

    # test member fn `__hash__`
    assert p1.__hash__() == p2.__hash__()
    assert p1.__hash__() != p3.__hash__()

    # test member fn `resize`
    p11 = p1.resize(
        og_img_h=10,
        og_img_w=10,
        new_img_h=100,
        new_img_w=100,
    )
    assert p11.x == p1.x * 10
    assert p11.y == p1.y * 10


def test_polygon():
    p1 = (-1, 0)
    p2 = (-5, 2)
    p3 = (-2, 5)
    coords = [p1, p2, p3, p1]

    # valid
    poly = Polygon([coords])
    poly_w_hole = Polygon([coords, coords])  # defines a hole

    # test validation
    with pytest.raises(ValueError):
        assert Polygon([[p1, p2, p3]])
    with pytest.raises(TypeError):
        Polygon(123)  # type: ignore - testing
    with pytest.raises(TypeError):
        Polygon([poly, 123])  # type: ignore - testing
    with pytest.raises(TypeError):
        Polygon([poly, [123]])  # type: ignore - testing

    # test property 'boundary'
    assert poly.boundary == coords
    assert poly_w_hole.boundary == coords

    # test property 'holes'
    assert poly.holes == []
    assert poly_w_hole.holes == [coords]

    # test property 'xmin'
    assert poly.xmin == -5

    # test property 'xmax'
    assert poly.xmax == -1

    # test property 'ymin'
    assert poly.ymin == 0

    # test property 'ymax'
    assert poly.ymax == 5


def test_box():
    p1 = (-1, -2)
    p2 = (10, -2)
    p3 = (10, 11)
    p4 = (-1, 11)
    coords = [[p1, p2, p3, p4, p1]]

    # test validation
    Box(coords)
    with pytest.raises(TypeError) as e:
        Box(polygon=p1)  # type: ignore - testing
    with pytest.raises(ValueError) as e:
        Box([[p1, p2, p3, p4]])
    assert "at least 4 points with the first point being repeated" in str(e)

    # test classmethod `from_extrema`
    assert (
        Box.from_extrema(xmin=-1, xmax=10, ymin=-2, ymax=11).get_value()
        == coords
    )


def test_multipolygon():
    p1 = (0, 0)
    p2 = (5, 0)
    p3 = (5, 5)
    p4 = (0, 5)
    coords = [p1, p2, p3, p4, p1]

    # valid
    MultiPolygon([[coords]])

    # test validation
    with pytest.raises(TypeError):
        MultiPolygon(coords)  # type: ignore - testing
    with pytest.raises(TypeError):
        MultiPolygon([coords])  # type: ignore - testing
    with pytest.raises(TypeError):
        MultiPolygon([[coords], 123])  # type: ignore - testing
    with pytest.raises(ValueError):
        MultiPolygon([[[coords]]])  # type: ignore - testing


def test_raster(raster_raw_mask):
    mask1 = np.ones((10, 10)) == 1
    poly1 = Polygon([[(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]])
    multipoly1 = MultiPolygon([[[(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]]])

    # valid
    Raster({"mask": mask1, "geometry": None})
    Raster({"mask": mask1, "geometry": poly1.get_value()})
    Raster({"mask": mask1, "geometry": multipoly1.get_value()})
    Raster.from_numpy(mask=mask1)
    Raster.from_geometry(geometry=poly1, height=10, width=10)
    Raster.from_geometry(geometry=multipoly1, height=10, width=10)

    # test validation
    with pytest.raises(TypeError):
        assert Raster({"mask": "test", "geometry": None})  # type: ignore - testing
    with pytest.raises(TypeError) as e:
        assert Raster(123)  # type: ignore - testing

    # test classmethod `from_numpy`
    mask2 = np.ones((10, 10, 10)) == 1
    mask3 = np.ones((10, 10))
    with pytest.raises(ValueError) as e:
        Raster.from_numpy(mask2)
    assert "raster only supports 2d arrays" in str(e)
    with pytest.raises(ValueError) as e:
        Raster.from_numpy(mask3)
    assert "Expecting a binary mask" in str(e)

    # test member fn `to_numpy`
    r = Raster.from_numpy(raster_raw_mask)
    value = r.encode_value()
    assert value
    assert (
        value["mask"]
        == "iVBORw0KGgoAAAANSUhEUgAAABQAAAAUAQAAAACl8iCgAAAAF0lEQVR4nGP4f4CBiYGBIGZgsP9AjDoAuysDE0GVDN8AAAAASUVORK5CYII="
    )
    assert (r.array == raster_raw_mask).all()



import datetime

import pytest

from valor.schemas import Dictionary


def test_validate_metadata():
    Dictionary({"test": "test"})
    Dictionary({"test": 1})
    Dictionary({"test": 1.0})

    with pytest.raises(TypeError):
        Dictionary({123: 123})  # type: ignore - testing

    # Check int to float conversion
    assert type(Dictionary({"test": 1})["test"]) is int
    assert type(Dictionary({"test": 1.0})["test"]) is float


def test_init_dictionary_from_builtin_dict():
    metadata = dict()
    metadata["a"] = int(123.4)
    metadata["b"] = float(123.4)
    metadata["c"] = str(123.4)
    metadata["d"] = datetime.datetime.fromisoformat("2023-01-01T12:12:12")
    metadata["e"] = datetime.date.fromisoformat("2023-01-01")
    metadata["f"] = datetime.time.fromisoformat("12:12:12:100000")
    metadata["g"] = datetime.timedelta(days=1)
    assert Dictionary(metadata).get_value() == metadata


def test_dump_metadata_to_json():
    metadata = dict()
    metadata["a"] = int(123.4)
    metadata["b"] = float(123.4)
    metadata["c"] = str(123.4)
    metadata["d"] = datetime.datetime.fromisoformat("2023-01-01T12:12:12")
    metadata["e"] = datetime.date.fromisoformat("2023-01-01")
    metadata["f"] = datetime.time.fromisoformat("12:12:12:100000")
    metadata["g"] = datetime.timedelta(days=1)
    assert Dictionary(metadata).get_value() == metadata

    assert Dictionary(metadata).to_dict() == {
        "type": "dictionary",
        "value": {
            "a": 123,
            "b": 123.4,
            "c": "123.4",
            "d": {"type": "datetime", "value": "2023-01-01T12:12:12"},
            "e": {"type": "date", "value": "2023-01-01"},
            "f": {"type": "time", "value": "12:12:12.100000"},
            "g": {"type": "duration", "value": 86400.0},
        },
    }


def test_dictionary_encoding():
    metadata = dict()
    metadata["a"] = int(123.4)
    metadata["b"] = float(123.4)
    metadata["c"] = str(123.4)
    metadata["d"] = datetime.datetime.fromisoformat("2023-01-01T12:12:12")
    metadata["e"] = datetime.date.fromisoformat("2023-01-01")
    metadata["f"] = datetime.time.fromisoformat("12:12:12:100000")
    metadata["g"] = datetime.timedelta(days=1)

    metadata_json = {
        "a": 123,
        "b": 123.4,
        "c": "123.4",
        "d": {"type": "datetime", "value": "2023-01-01T12:12:12"},
        "e": {"type": "date", "value": "2023-01-01"},
        "f": {"type": "time", "value": "12:12:12.100000"},
        "g": {"type": "duration", "value": 86400.0},
    }

    assert Dictionary(metadata).encode_value() == metadata_json
    assert Dictionary.decode_value(metadata_json).get_value() == metadata



from typing import Any, Dict, List, Optional, Union

import numpy as np

from valor.schemas.symbolic.types import (
    Boolean,
    Box,
    ContextList,
    Dictionary,
    Embedding,
    Equatable,
    Float,
)
from valor.schemas.symbolic.types import List as SymbolicList
from valor.schemas.symbolic.types import (
    Polygon,
    Raster,
    String,
    Variable,
    _convert_simple_variables_to_standard_types,
    get_type_by_name,
)


def _get_schema_type_by_name(name: str):
    types_ = {
        "label": Label,
        "annotation": Annotation,
        "datum": Datum,
    }
    return get_type_by_name(name=name, additional_types=types_)


class StaticCollection(Equatable):
    """
    A static collection is a Variable that defines its contents by static attributes.
    """

    def __init__(self, **kwargs):
        if set(kwargs.keys()) != set(self._get_static_types().keys()):
            kwarg_keys = set(kwargs.keys())
            static_keys = set(self._get_static_types().keys())
            raise ValueError(
                f"Expected the following keyword arguments '{static_keys}'. Received '{kwarg_keys}'."
            )
        for k, v in kwargs.items():
            setattr(self, k, v)
        super().__init__(value=None)

    @classmethod
    def nullable(cls, *args, **kwargs):
        """
        Initializes variable with an optional value.
        """
        raise NotImplementedError(
            "Static collections do not define 'nullable'."
        )

    @classmethod
    def symbolic(
        cls,
        name: Optional[str] = None,
        key: Optional[str] = None,
        attribute: Optional[str] = None,
        owner: Optional[str] = None,
    ):
        """
        Initializes the object and its attributes as symbols.

        Parameters
        ----------
        name : str, optional
            The name of the symbol.
        key : str, optional
            The key of the value if its a dictionary element.
        attribute : str, optional
            The name of a an attribute this symbol represents.
        owner : str, optional
            The name of an object that this symbol belongs to.
        """
        obj = super().symbolic(name, key, attribute, owner)
        for __name, __type in obj._get_static_types().items():
            if not issubclass(__type, Variable):
                raise TypeError
            setattr(
                obj,
                __name,
                __type.symbolic(owner=cls.__name__.lower(), name=__name),
            )
        return obj

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return dict()

    def format(self, __name: str, __value: Any) -> Any:
        """Either formats or passes throught a name-value pair."""
        if __name in self._get_static_types():
            __type = self._get_static_types()[__name]
            if not isinstance(__value, __type):
                __fmt = (
                    self.formatting()[__name]
                    if __name in self.formatting()
                    else __type
                )
                if issubclass(__type, StaticCollection):
                    return __fmt(**__value)
                else:
                    return __fmt(__value)
        return __value

    def __setattr__(self, __name: str, __value: Any) -> None:
        super().__setattr__(__name, self.format(__name, __value))

    def __getattribute__(self, __name: str) -> Any:
        ret = super().__getattribute__(__name)
        if isinstance(ret, Variable) and ret.is_value:
            return _convert_simple_variables_to_standard_types(ret)
        return ret

    @classmethod
    def __validate__(cls, value: Any):
        """Validate typing."""
        if value is not None:
            raise TypeError(
                "A StaticCollection does not store an internal value."
            )

    @classmethod
    def decode_value(cls, value: dict):
        """Decode object from JSON compatible dictionary."""
        kwargs = dict()
        types = cls._get_static_types()
        for k, v in value.items():
            type_ = types.get(k)
            if type_ and issubclass(type_, Variable):
                kwargs[k] = type_.decode_value(v)
            else:
                kwargs[k] = v
        return cls(**kwargs)

    def encode_value(self):
        """Encode object to JSON compatible dictionary."""
        return {
            k: (v.encode_value() if hasattr(v, "encode_value") else v)
            for k, v in self._get_dynamic_values().items()
        }

    @classmethod
    def _get_static_types(cls) -> Dict[str, type]:
        """Returns any static members that inherit from 'Variable'."""
        fields = getattr(cls, "__annotations__", dict())
        retval = dict()
        for k, v in fields.items():
            if isinstance(v, type) and issubclass(v, Variable):
                retval[k] = v
            elif isinstance(v, str):
                retval[k] = _get_schema_type_by_name(v)
            else:
                raise NotImplementedError(
                    f"Unknown typing. Attribute '{k}' with type '{v}'."
                )
        return retval

    def _get_dynamic_values(self):
        """Returns the values of attributes that inherit from 'Variable'."""
        return {
            name: self.__getattribute__(name)
            for name in self._get_static_types().keys()
        }

    def __repr__(self):
        if self.is_symbolic:
            return super().__repr__()
        return f"{self.__class__.__name__}({self.encode_value().__repr__()})"

    def __str__(self):
        if self.is_symbolic:
            return super().__str__()
        return str(self.encode_value())


class Label(StaticCollection):
    """
    An object for labeling datasets, models, and annotations.

    Attributes
    ----------
    key : String
        The class label key.
    value : String
        The class label value.
    score : Score
        The label score.

    Examples
    --------
    >>> Label(key="k1", value="v1")
    >>> Label(key="k1", value="v1", score=None)
    >>> Label(key="k1", value="v1", score=0.9)
    """

    key: String = String.symbolic(owner="label", name="key")
    value: String = String.symbolic(owner="label", name="value")
    score: Float = Float.symbolic(owner="label", name="score")

    def __init__(
        self,
        *,
        key: str,
        value: str,
        score: Union[float, np.floating, None] = None,
    ):
        """
        Initializes an instance of a label.

        Attributes
        ----------
        key : str
            The class label key.
        value : str
            The class label value.
        score : float, optional
            The label score.
        """
        super().__init__(key=key, value=value, score=score)

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return {
            "score": Float.nullable,
        }

    def tuple(self):
        """
        Defines how the `Label` is turned into a tuple.

        Returns
        ----------
        tuple
            A tuple of the `Label's` arguments.
        """
        return (self.key, self.value, self.score)


class Annotation(StaticCollection):
    """
    A class used to annotate `GroundTruths` and `Predictions`.

    Attributes
    ----------
    metadata: Dictionary
        A dictionary of metadata that describes the `Annotation`.
    labels: List[Label], optional
        A list of labels to use for the `Annotation`.
    bounding_box: Box
        A bounding box to assign to the `Annotation`.
    polygon: BoundingPolygon
        A polygon to assign to the `Annotation`.
    raster: Raster
        A raster to assign to the `Annotation`.
    embedding: List[float]
        An embedding, described by a list of values with type float and a maximum length of 16,000.
    text: str, optional
        A piece of text to assign to the `Annotation`.
    context_list: List[str], optional
        A list of contexts associated with an `Annotation`.
    is_instance: bool, optional
        A boolean describing whether we should treat the Raster attached to an annotation as an instance segmentation or not. If set to true, then the Annotation will be validated for use in object detection tasks. If set to false, then the Annotation will be validated for use in semantic segmentation tasks.
    implied_task_types: list[str], optional
        The validated task types that are applicable to each Annotation. Doesn't need to bet set by the user.

    Examples
    --------

    Classification
    >>> Annotation.create(
    ...     labels=[
    ...         Label(key="class", value="dog"),
    ...         Label(key="category", value="animal"),
    ...     ]
    ... )

    Object-Detection Box
    >>> annotation = Annotation(
    ...     labels=[Label(key="k1", value="v1")],
    ...     bounding_box=box2,
    ... )

    Object-Detection Polygon
    >>> annotation = Annotation(
    ...     labels=[Label(key="k1", value="v1")],
    ...     polygon=BoundingPolygon(...),
    ... )

     Raster
    >>> annotation = Annotation(
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=Raster(...),
    ...     is_instance=True
    ... )

    Object-Detection with all supported Geometries defined.
    >>> Annotation(
    ...     labels=[Label(key="k1", value="v1")],
    ...     bounding_box=Box(...),
    ...     polygon=BoundingPolygon(...),
    ...     raster=Raster(...),
    ...     is_instance=True,
    ... )

    Semantic-Segmentation Raster
    >>> annotation = Annotation(
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=Raster(...),
    ...     is_instance=False # or None
    ... )

    Text Generation
    >>> annotation = Annotation(
    ...     text="Yes, Lincoln won the election of 1860. He received the highest number of votes...",
    ...     context_list=["Republican speakers focused first on...", "Lincoln received 1,866,452 votes...", ...],
    ... )
    """

    metadata: Dictionary = Dictionary.symbolic(
        owner="annotation", name="metadata"
    )
    labels: SymbolicList[Label] = SymbolicList[Label].symbolic(
        owner="annotation", name="labels"
    )
    bounding_box: Box = Box.symbolic(owner="annotation", name="bounding_box")
    polygon: Polygon = Polygon.symbolic(owner="annotation", name="polygon")
    raster: Raster = Raster.symbolic(owner="annotation", name="raster")
    embedding: Embedding = Embedding.symbolic(
        owner="annotation", name="embedding"
    )
    text: String = String.symbolic(owner="annotation", name="text")
    context_list: ContextList = ContextList.symbolic(
        owner="annotation", name="context_list"
    )
    is_instance: Boolean = Boolean.symbolic(
        owner="annotation", name="is_instance"
    )
    implied_task_types: SymbolicList[String] = SymbolicList[String].symbolic(
        owner="annotation", name="implied_task_types"
    )

    def __init__(
        self,
        *,
        metadata: Optional[dict] = None,
        labels: Optional[List[Label]] = None,
        bounding_box: Optional[Box] = None,
        polygon: Optional[Polygon] = None,
        raster: Optional[Raster] = None,
        embedding: Optional[Embedding] = None,
        text: Optional[str] = None,
        context_list: Optional[List[str]] = None,
        is_instance: Optional[bool] = None,
        implied_task_types: Optional[List[String]] = None,
    ):
        """
        Constructs an annotation.

        Parameters
        ----------
        metadata: Dict[str, Union[int, float, str, bool, datetime.datetime, datetime.date, datetime.time]]
            A dictionary of metadata that describes the `Annotation`.
        labels: List[Label]
            A list of labels to use for the `Annotation`.
        bounding_box: Box, optional
            A bounding box annotation.
        polygon: Polygon, optional
            A polygon annotation.
        raster: Raster, optional
            A raster annotation.
        embedding: List[float], optional
            An embedding, described by a list of values with type float and a maximum length of 16,000.
        text: str, optional
            A text annotation.
        context_list: List[str], optional
            A list of contexts associated to the annotation text. Not all text annotations will have context_list.
        is_instance: bool, optional
            A boolean describing whether we should treat the Raster attached to an annotation as an instance segmentation or not. If set to true, then the Annotation will be validated for use in object detection tasks. If set to false, then the Annotation will be validated for use in semantic segmentation tasks.
        implied_task_types: list[str], optional
            The validated task types that are applicable to each Annotation. Doesn't need to bet set by the user.
        """
        super().__init__(
            metadata=metadata if metadata else dict(),
            labels=labels if labels else list(),
            bounding_box=bounding_box,
            polygon=polygon,
            raster=raster,
            embedding=embedding,
            text=text,
            context_list=context_list,
            is_instance=is_instance,
            implied_task_types=implied_task_types,
        )

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return {
            "bounding_box": Box.nullable,
            "polygon": Polygon.nullable,
            "raster": Raster.nullable,
            "embedding": Embedding.nullable,
            "text": String.nullable,
            "context_list": ContextList.nullable,
            "is_instance": Boolean.nullable,
            "implied_task_types": SymbolicList,
        }


class Datum(StaticCollection):
    """
    A class used to store information about a datum for either a 'GroundTruth' or a 'Prediction'.

    Attributes
    ----------
    uid : String
        The UID of the datum.
    text : String, optional
        The text of the datum, if the datum is a piece of text, otherwise None.
    metadata : Dictionary
        A dictionary of metadata that describes the datum.

    Examples
    --------
    >>> Datum(uid="uid1")
    >>> Datum(uid="uid1", metadata={})
    >>> Datum(uid="uid1", metadata={"foo": "bar", "pi": 3.14})
    >>> Datum(uid="uid2", text="Did Lincoln win the election of 1860?", metadata={"query_created_by": "Alice"})
    """

    uid: String = String.symbolic(owner="datum", name="uid")
    text: String = String.symbolic(owner="datum", name="text")
    metadata: Dictionary = Dictionary.symbolic(owner="datum", name="metadata")

    def __init__(
        self,
        *,
        uid: str,
        text: Optional[str] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Constructs a datum.

        Parameters
        ----------
        uid : str
            The UID of the datum.
        text : str, optional
            The text of the datum, if the datum is a piece of text, otherwise None.
        metadata : dict, optional
            A dictionary of metadata that describes the datum.
        """
        super().__init__(
            uid=uid, text=text, metadata=metadata if metadata else dict()
        )

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return {
            "text": String.nullable,
        }



from __future__ import annotations

import datetime
import json
import os
import time
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from valor.client import ClientConnection, connect, get_connection
from valor.enums import (
    AnnotationType,
    EvaluationStatus,
    MetricType,
    TableStatus,
    TaskType,
)
from valor.exceptions import (
    ClientException,
    DatasetDoesNotExistError,
    EvaluationDoesNotExist,
    ModelDoesNotExistError,
)
from valor.schemas import (
    Annotation,
    Datum,
    Dictionary,
    EvaluationParameters,
    EvaluationRequest,
    Filter,
    Label,
)
from valor.schemas import List as SymbolicList
from valor.schemas import StaticCollection, String


class GroundTruth(StaticCollection):
    """
    An object describing a ground truth (e.g., a human-drawn bounding box on an image).

    Attributes
    ----------
    datum : Datum
        The datum associated with the groundtruth.
    annotations : List[Annotation]
        The list of annotations associated with the groundtruth.

    Examples
    --------
    >>> GroundTruth(
    ...     datum=Datum(uid="uid1"),
    ...     annotations=[
    ...         Annotation(
    ...             labels=[Label(key="k1", value="v1")],
    ...         )
    ...     ]
    ... )
    """

    datum: Datum = Datum.symbolic(owner="groundtruth", name="datum")
    annotations: SymbolicList[Annotation] = SymbolicList[Annotation].symbolic(
        owner="groundtruth", name="annotations"
    )

    def __init__(
        self,
        *,
        datum: Datum,
        annotations: List[Annotation],
    ):
        """
        Creates a ground truth.

        Parameters
        ----------
        datum : Datum
            The datum that the ground truth is operating over.
        annotations : List[Annotation]
            The list of ground truth annotations.
        """
        super().__init__(datum=datum, annotations=annotations)

        for annotation in self.annotations:
            for label in annotation.labels:
                if label.score is not None:
                    raise ValueError(
                        "GroundTruth labels should not have scores."
                    )


class Prediction(StaticCollection):
    """
    An object describing a prediction (e.g., a machine-drawn bounding box on an image).

    Attributes
    ----------
    datum : Datum
        The datum associated with the prediction.
    annotations : List[Annotation]
        The list of annotations associated with the prediction.

    Examples
    --------
    >>> Prediction(
    ...     datum=Datum(uid="uid1"),
    ...     annotations=[
    ...         Annotation(
    ...             labels=[
    ...                 Label(key="k1", value="v1", score=0.9),
    ...                 Label(key="k1", value="v1", score=0.1)
    ...             ],
    ...         )
    ...     ]
    ... )
    """

    datum: Datum = Datum.symbolic(owner="prediction", name="datum")
    annotations: SymbolicList[Annotation] = SymbolicList[Annotation].symbolic(
        owner="prediction", name="annotations"
    )

    def __init__(
        self,
        *,
        datum: Datum,
        annotations: List[Annotation],
    ):
        """
        Creates a prediction.

        Parameters
        ----------
        datum : Datum
            The datum that the prediction is operating over.
        annotations : List[Annotation]
            The list of predicted annotations.
        """
        super().__init__(datum=datum, annotations=annotations)


class Evaluation:
    """
    Wraps `valor.client.Job` to provide evaluation-specifc members.
    """

    def __init__(
        self, connection: Optional[ClientConnection] = None, **kwargs
    ):
        """
        Defines important attributes of the API's `EvaluationResult`.

        Attributes
        ----------
        id : int
            The ID of the evaluation.
        dataset_names : list[str]
            The names of the datasets the model was evaluated over.
        model_name : str
            The name of the evaluated model.
        filters : dict
            The filter used to select data partitions for evaluation.
        status : EvaluationStatus
            The status of the evaluation.
        metrics : List[dict]
            A list of metric dictionaries returned by the job.
        confusion_matrices : List[dict]
            A list of confusion matrix dictionaries returned by the job.
        meta : dict[str, str | float | dict], optional
            A dictionary of metadata describing the evaluation run.
        """
        if not connection:
            connection = get_connection()
        self.conn = connection
        self.update(**kwargs)

    def update(
        self,
        *_,
        id: int,
        dataset_names: list[str],
        model_name: str,
        filters: dict,
        parameters: EvaluationParameters,
        status: EvaluationStatus,
        metrics: List[Dict],
        confusion_matrices: List[Dict],
        created_at: str,
        meta: dict[str, str | float | dict] | None,
        **kwargs,
    ):
        self.id = id
        self.dataset_names = dataset_names
        self.model_name = model_name
        self.filters = filters
        self.parameters = (
            EvaluationParameters(**parameters)
            if isinstance(parameters, dict)
            else parameters
        )
        self.status = EvaluationStatus(status)
        self.metrics = metrics
        self.meta = meta
        self.confusion_matrices = confusion_matrices
        self.kwargs = kwargs
        self.ignored_pred_labels: Optional[List[Label]] = None
        self.missing_pred_labels: Optional[List[Label]] = None
        self.created_at = datetime.datetime.strptime(
            created_at, "%Y-%m-%dT%H:%M:%S.%fZ"
        ).replace(tzinfo=datetime.timezone.utc)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def poll(self) -> EvaluationStatus:
        """
        Poll the back end.

        Updates the evaluation with the latest state from the back end.

        Returns
        -------
        enums.EvaluationStatus
            The status of the evaluation.

        Raises
        ----------
        ClientException
            If an Evaluation with the given `evaluation_id` is not found.
        """
        response = self.conn.get_evaluations(
            evaluation_ids=[self.id],
            models=None,
            datasets=None,
            metrics_to_sort_by=None,
        )
        if not response:
            raise EvaluationDoesNotExist(self.id)
        self.update(**response[0])
        return self.status

    def wait_for_completion(
        self,
        *,
        timeout: Optional[int] = None,
        interval: float = 1.0,
    ) -> EvaluationStatus:
        """
        Blocking function that waits for evaluation to finish.

        Parameters
        ----------
        timeout : int, optional
            Length of timeout in seconds.
        interval : float, default=1.0
            Polling interval in seconds.
        """
        t_start = time.time()
        while self.poll() not in [
            EvaluationStatus.DONE,
            EvaluationStatus.FAILED,
        ]:
            time.sleep(interval)
            if timeout and time.time() - t_start > timeout:
                raise TimeoutError
        return self.status

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        return json.dumps(self.to_dict(), indent=4)

    def to_dict(self) -> dict:
        """
        Defines how a `valor.Evaluation` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing an evaluation.
        """
        return {
            "id": self.id,
            "dataset_names": self.dataset_names,
            "model_name": self.model_name,
            "filters": self.filters,
            "parameters": asdict(self.parameters),
            "status": self.status.value,
            "metrics": self.metrics,
            "confusion_matrices": self.confusion_matrices,
            "meta": self.meta,
            **self.kwargs,
        }

    def to_dataframe(
        self,
        stratify_by: Optional[Tuple[str, str]] = None,
    ):
        """
        Get all metrics associated with a Model and return them in a `pd.DataFrame`.

        Returns
        ----------
        pd.DataFrame
            Evaluation metrics being displayed in a `pd.DataFrame`.

        Raises
        ------
        ModuleNotFoundError
            This function requires the use of `pandas.DataFrame`.

        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Must have pandas installed to use `get_metric_dataframes`."
            )

        if not stratify_by:
            column_type = "evaluation"
            column_name = self.id
        else:
            column_type = stratify_by[0]
            column_name = stratify_by[1]

        metrics = [
            {**metric, column_type: column_name} for metric in self.metrics
        ]
        df = pd.DataFrame(metrics)
        for k in ["label", "parameters"]:
            df[k] = df[k].fillna("n/a")
        df["parameters"] = df["parameters"].apply(json.dumps)
        df["label"] = df["label"].apply(
            lambda x: f"{x['key']}: {x['value']}" if x != "n/a" else x
        )
        df = df.pivot(
            index=["type", "parameters", "label"], columns=[column_type]
        )
        return df


@dataclass
class DatasetSummary:
    """Dataclass for storing dataset summary information"""

    name: str
    num_datums: int
    num_annotations: int
    num_bounding_boxes: int
    num_polygons: int
    num_rasters: int
    task_types: list[list[str]]
    labels: List[Label]
    datum_metadata: List[dict]
    annotation_metadata: List[dict]

    def __post_init__(self):
        for i, label in enumerate(self.labels):
            if isinstance(label, dict):
                self.labels[i] = Label(**label)


class Dataset(StaticCollection):
    """
    A class describing a given dataset.

    Attributes
    ----------
    name : String
        The name of the dataset.
    metadata : Dictionary
        A dictionary of metadata that describes the dataset.

    Examples
    --------
    >>> Dataset.create(name="dataset1")
    >>> Dataset.create(name="dataset1", metadata={})
    >>> Dataset.create(name="dataset1", metadata={"foo": "bar", "pi": 3.14})
    """

    name: String = String.symbolic(owner="dataset", name="name")
    metadata: Dictionary = Dictionary.symbolic(
        owner="dataset", name="metadata"
    )

    def __init__(
        self,
        *,
        name: str,
        metadata: Optional[dict] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Creates a local instance of a dataset.

        Use 'Dataset.create' classmethod to create a dataset with persistence.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict, optional
            A dictionary of metadata that describes the dataset.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        self.conn = connection
        super().__init__(name=name, metadata=metadata if metadata else dict())

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
        connection: Optional[ClientConnection] = None,
    ) -> Dataset:
        """
        Creates a dataset that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict, optional
            A dictionary of metadata that describes the dataset.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        dataset = cls(name=name, metadata=metadata, connection=connection)
        Client(dataset.conn).create_dataset(dataset)
        return dataset

    @classmethod
    def get(
        cls,
        name: str,
        connection: Optional[ClientConnection] = None,
    ) -> Union[Dataset, None]:
        """
        Retrieves a dataset from the back end database.

        Parameters
        ----------
        name : str
            The name of the dataset.

        Returns
        -------
        Union[valor.Dataset, None]
            The dataset or 'None' if it doesn't exist.
        """
        return Client(connection).get_dataset(name)

    def add_groundtruth(
        self,
        groundtruth: GroundTruth,
    ) -> None:
        """
        Add a ground truth to the dataset.

        Parameters
        ----------
        groundtruth : GroundTruth
            The ground truth to create.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=[groundtruth],
        )

    def add_groundtruths(
        self,
        groundtruths: List[GroundTruth],
        ignore_existing_datums: bool = False,
        timeout: Optional[float] = 10.0,
    ) -> None:
        """
        Add multiple ground truths to the dataset.

        Parameters
        ----------
        groundtruths : List[GroundTruth]
            The ground truths to create.
        ignore_existing_datums : bool, default=False
            If True, will ignore datums that already exist in the backend.
            If False, will raise an error if any datums already exist.
            Default is False.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=groundtruths,
            ignore_existing_datums=ignore_existing_datums,
            timeout=timeout,
        )

    def get_groundtruth(
        self,
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        datum : Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_groundtruth(dataset=self, datum=datum)

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given dataset.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the dataset.
        """
        return Client(self.conn).get_labels_from_dataset(self)

    def get_datums(self, filters: Optional[Filter] = None) -> List[Datum]:
        """
        Get all datums associated with a given dataset.

        Parameters
        ----------
        filters : Filter, optional
            An optional datum filter.

        Returns
        ----------
        List[Datum]
            A list of `Datums` associated with the dataset.
        """
        if filters is None:
            filters = Filter()
        filters.datasets = Dataset.name == self.name  # type: ignore - #issue 605
        return Client(self.conn).get_datums(filters=filters)

    def get_evaluations(
        self,
        metrics_to_sort_by: Optional[
            Dict[str, Union[Dict[str, str], str]]
        ] = None,
        *_,
        timeout: Optional[float] = None,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given dataset.

        Parameters
        ----------
        metrics_to_sort_by : dict[str, str | dict[str, str]], optional
            An optional dict of metric types to sort the evaluations by.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the dataset.
        """
        return Client(self.conn).get_evaluations(
            datasets=[self],
            metrics_to_sort_by=metrics_to_sort_by,
            timeout=timeout,
        )

    def get_summary(
        self, *_, timeout: Optional[float] = None
    ) -> DatasetSummary:
        """
        Get the summary of a given dataset.

        Parameters
        ----------
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        DatasetSummary
            The summary of the dataset. This class has the following fields:

            name : name of the dataset

            num_datums : total number of datums in the dataset

            num_annotations : total number of labeled annotations in the dataset; if an
            object (such as a bounding box) has multiple labels, then each label is counted separately

            num_bounding_boxes : total number of bounding boxes in the dataset

            num_polygons : total number of polygons in the dataset

            num_rasters : total number of rasters in the dataset

            labels : list of the unique labels in the dataset

            datum_metadata : list of the unique metadata dictionaries in the dataset that are associated
            to datums

            groundtruth_annotation_metadata : list of the unique metadata dictionaries in the dataset that are
            associated to annotations
        """
        return Client(self.conn).get_dataset_summary(self.name, timeout=timeout)  # type: ignore

    def finalize(
        self,
    ):
        """
        Finalizes the dataset such that new ground truths cannot be added to it.
        """
        return Client(self.conn).finalize_dataset(self)

    def delete(
        self,
        timeout: int = 0,
    ):
        """
        Delete the dataset from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_dataset(self.name, timeout)  # type: ignore


class Model(StaticCollection):
    """
    A class describing a model that was trained on a particular dataset.

    Attributes
    ----------
    name : String
        The name of the model.
    metadata : Dictionary
        A dictionary of metadata that describes the model.

    Examples
    --------
    >>> Model.create(name="model1")
    >>> Model.create(name="model1", metadata={})
    >>> Model.create(name="model1", metadata={"foo": "bar", "pi": 3.14})
    """

    name: String = String.symbolic(owner="model", name="name")
    metadata: Dictionary = Dictionary.symbolic(owner="model", name="metadata")

    def __init__(
        self,
        *,
        name: str,
        metadata: Optional[dict] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Creates a local instance of a model.

        Use 'Model.create' classmethod to create a model with persistence.

        Parameters
        ----------
        name : String
            The name of the model.
        metadata : Dictionary
            A dictionary of metadata that describes the model.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        self.conn = connection
        super().__init__(name=name, metadata=metadata if metadata else dict())

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
        connection: Optional[ClientConnection] = None,
        **_,
    ) -> Model:
        """
        Creates a model that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the model.
        metadata : dict, optional
            A dictionary of metadata that describes the model.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        model = cls(name=name, metadata=metadata, connection=connection)
        Client(connection).create_model(model)
        return model

    @classmethod
    def get(
        cls,
        name: str,
        connection: Optional[ClientConnection] = None,
    ) -> Union[Model, None]:
        """
        Retrieves a model from the back end database.

        Parameters
        ----------
        name : str
            The name of the model.
        connection : ClientConnnetion, optional
            An optional Valor client object for interacting with the API.

        Returns
        -------
        Union[valor.Model, None]
            The model or 'None' if it doesn't exist.
        """
        return Client(connection).get_model(name)

    def add_prediction(
        self,
        dataset: Dataset,
        prediction: Prediction,
    ) -> None:
        """
        Add a prediction to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        prediction : valor.Prediction
            The prediction to create.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=[prediction],
        )

    def add_predictions(
        self,
        dataset: Dataset,
        predictions: List[Prediction],
        timeout: Optional[float] = 10.0,
    ) -> None:
        """
        Add multiple predictions to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        predictions : List[valor.Prediction]
            The predictions to create.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=predictions,
            timeout=timeout,
        )

    def get_prediction(
        self, dataset: Union[Dataset, str], datum: Union[Datum, str]
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset : Union[Dataset, str]
            The dataset the datum belongs to.
        datum : Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_prediction(
            dataset=dataset, model=self, datum=datum
        )

    def finalize_inferences(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes the model over a dataset such that new predictions cannot be added to it.
        """
        return Client(self.conn).finalize_inferences(
            dataset=dataset, model=self
        )

    def _create_label_map(
        self,
        label_map: Optional[Dict[Label, Label]],
    ) -> Union[List[List[List[str]]], None]:
        """Convert a dictionary of label maps to a serializable list format."""
        if not label_map:
            return None

        if not isinstance(label_map, dict) or not all(
            [
                isinstance(key, Label) and isinstance(value, Label)
                for key, value in label_map.items()
            ]
        ):
            raise TypeError(
                "label_map should be a dictionary with valid Labels for both the key and value."
            )

        return_value = []
        for key, value in label_map.items():
            if not all(
                [
                    (isinstance(v.key, str) and isinstance(v.value, str))
                    for v in [key, value]
                ]
            ):
                raise TypeError
            return_value.append(
                [
                    [key.key, key.value],
                    [value.key, value.value],
                ]
            )
        return return_value

    def evaluate_classification(
        self,
        datasets: Union[Dataset, List[Dataset]],
        filters: Optional[Filter] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        pr_curve_max_examples: int = 1,
        metrics_to_return: Optional[List[MetricType]] = None,
        *_,
        allow_retries: bool = False,
        timeout: Optional[float] = None,
    ) -> Evaluation:
        """
        Start a classification evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]]
            The dataset or list of datasets to evaluate against.
        filters : Filter, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        metrics_to_return : List[MetricType], optional
            The list of metrics to compute, store, and return to the user.
        allow_retries : bool, default = False
            Option to retry previously failed evaluations.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if metrics_to_return and not set(metrics_to_return).issubset(
            MetricType.classification()
        ):
            raise ValueError(
                f"The following metrics are not supported for classification: '{set(metrics_to_return) - MetricType.classification()}'"
            )

        # format request
        datasets = datasets if isinstance(datasets, list) else [datasets]
        filters = filters if filters else Filter()
        request = EvaluationRequest(
            dataset_names=[dataset.name for dataset in datasets],  # type: ignore - issue #604
            model_names=[self.name],  # type: ignore - issue #604
            filters=filters,
            parameters=EvaluationParameters(
                task_type=TaskType.CLASSIFICATION,
                label_map=self._create_label_map(label_map=label_map),
                pr_curve_max_examples=pr_curve_max_examples,
                metrics_to_return=metrics_to_return,
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(
            request, allow_retries=allow_retries, timeout=timeout
        )
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_detection(
        self,
        datasets: Union[Dataset, List[Dataset]],
        filters: Optional[Filter] = None,
        convert_annotations_to_type: Optional[AnnotationType] = None,
        iou_thresholds_to_compute: Optional[List[float]] = None,
        iou_thresholds_to_return: Optional[List[float]] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        recall_score_threshold: float = 0,
        metrics_to_return: Optional[List[MetricType]] = None,
        pr_curve_iou_threshold: float = 0.5,
        pr_curve_max_examples: int = 1,
        *_,
        allow_retries: bool = False,
        timeout: Optional[float] = None,
    ) -> Evaluation:
        """
        Start an object-detection evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]]
            The dataset or list of datasets to evaluate against.
        filters : Filter, optional
            Optional set of constraints to filter evaluation by.
        convert_annotations_to_type : enums.AnnotationType, optional
            Forces the object detection evaluation to compute over this type.
        iou_thresholds_to_compute : List[float], optional
            Thresholds to compute mAP against.
        iou_thresholds_to_return : List[float], optional
            Thresholds to return AP for. Must be subset of `iou_thresholds_to_compute`.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        recall_score_threshold : float, default=0
            The confidence score threshold for use when determining whether to count a prediction as a true positive or not while calculating Average Recall.
        metrics_to_return : List[MetricType], optional
            The list of metrics to compute, store, and return to the user.
        pr_curve_iou_threshold : float, optional
            The IOU threshold to use when calculating precision-recall curves. Defaults to 0.5.
        pr_curve_max_examples : int, optional
            The maximum number of datum examples to store when calculating PR curves.
        allow_retries : bool, default = False
            Option to retry previously failed evaluations.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if metrics_to_return and not set(metrics_to_return).issubset(
            MetricType.object_detection()
        ):
            raise ValueError(
                f"The following metrics are not supported for object detection: '{set(metrics_to_return) - MetricType.object_detection()}'"
            )

        if iou_thresholds_to_compute is None:
            iou_thresholds_to_compute = [
                round(0.5 + 0.05 * i, 2) for i in range(10)
            ]
        if iou_thresholds_to_return is None:
            iou_thresholds_to_return = [0.5, 0.75]

        # format request
        parameters = EvaluationParameters(
            task_type=TaskType.OBJECT_DETECTION,
            convert_annotations_to_type=convert_annotations_to_type,
            iou_thresholds_to_compute=iou_thresholds_to_compute,
            iou_thresholds_to_return=iou_thresholds_to_return,
            label_map=self._create_label_map(label_map=label_map),
            recall_score_threshold=recall_score_threshold,
            metrics_to_return=metrics_to_return,
            pr_curve_iou_threshold=pr_curve_iou_threshold,
            pr_curve_max_examples=pr_curve_max_examples,
        )
        datasets = datasets if isinstance(datasets, list) else [datasets]
        filters = filters if filters else Filter()
        request = EvaluationRequest(
            dataset_names=[dataset.name for dataset in datasets],  # type: ignore - issue #604
            model_names=[self.name],  # type: ignore - issue #604
            filters=filters,
            parameters=parameters,
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(
            request, allow_retries=allow_retries, timeout=timeout
        )
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_segmentation(
        self,
        datasets: Union[Dataset, List[Dataset]],
        filters: Optional[Filter] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        metrics_to_return: Optional[List[MetricType]] = None,
        *_,
        allow_retries: bool = False,
        timeout: Optional[float] = None,
    ) -> Evaluation:
        """
        Start a semantic-segmentation evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]]
            The dataset or list of datasets to evaluate against.
        filters : Filter, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        metrics_to_return : List[MetricType], optional
            The list of metrics to compute, store, and return to the user.
        allow_retries : bool, default = False
            Option to retry previously failed evaluations.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion
        """
        if metrics_to_return and not set(metrics_to_return).issubset(
            MetricType.semantic_segmentation()
        ):
            raise ValueError(
                f"The following metrics are not supported for semantic segmentation: '{set(metrics_to_return) - MetricType.semantic_segmentation()}'"
            )

        # format request
        datasets = datasets if isinstance(datasets, list) else [datasets]
        filters = filters if filters else Filter()
        request = EvaluationRequest(
            dataset_names=[dataset.name for dataset in datasets],  # type: ignore - issue #604
            model_names=[self.name],  # type: ignore - issue #604
            filters=filters,
            parameters=EvaluationParameters(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                label_map=self._create_label_map(label_map=label_map),
                metrics_to_return=metrics_to_return,
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(
            request, allow_retries=allow_retries, timeout=timeout
        )
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_text_generation(
        self,
        datasets: Union[Dataset, List[Dataset]],
        metrics_to_return: List[MetricType],
        filters: Optional[Filter] = None,
        llm_api_params: Optional[Dict[str, Union[str, dict]]] = None,
        metric_params: Optional[Dict[MetricType, Dict[str, Any]]] = None,
    ) -> Evaluation:
        """
        Start a classification evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]]
            The dataset or list of datasets to evaluate against.
        metrics_to_return : List[MetricType]
            The list of metrics to compute, store, and return to the user. This is not optional for text generation evaluations.
        filters : Filter, optional
            Optional set of constraints to filter evaluation by.
        llm_api_params : Dict[str, Union[str,dict]], optional
            A dictionary of parameters for the LLM API.
        metric_params : Dict[MetricType, Dict[str,Any]], optional
            A dictionary of parameters for the metrics used in the evaluation. The keys should be the metrics and the values should be dictionaries of parameters for those metrics.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if not set(metrics_to_return).issubset(MetricType.text_generation()):
            raise ValueError(
                f"The following metrics are not supported for text generation: '{set(metrics_to_return) - MetricType.text_generation()}'"
            )

        # If no api_key is provided, check the environment variables for an api key.
        if llm_api_params is not None:
            if "api_key" not in llm_api_params:
                if "client" not in llm_api_params:
                    raise ValueError(
                        "The client must be specified in the llm_api_params."
                    )

                if llm_api_params["client"] == "openai":
                    api_key = os.getenv("OPENAI_API_KEY", None)
                elif llm_api_params["client"] == "mistral":
                    api_key = os.getenv("MISTRAL_API_KEY", None)
                elif llm_api_params["client"] == "mock":
                    api_key = ""
                else:
                    raise ValueError(
                        "The client specified in llm_api_params is not supported."
                    )

                if api_key is not None:
                    llm_api_params["api_key"] = api_key

        bleu_weights = None
        rouge_types = None
        rouge_use_stemmer = None
        if metric_params is not None:
            if not all(
                metric in metrics_to_return for metric in metric_params.keys()
            ):
                raise ValueError(
                    "All metrics in metric_params must be in metrics_to_return."
                )

            if MetricType.BLEU in metric_params:
                bleu_weights = metric_params[MetricType.BLEU].get("weights")

            if MetricType.ROUGE in metric_params:
                rouge_types = metric_params[MetricType.ROUGE].get(
                    "rouge_types"
                )
                rouge_use_stemmer = metric_params[MetricType.ROUGE].get(
                    "use_stemmer"
                )

        # format request
        datasets = datasets if isinstance(datasets, list) else [datasets]
        filters = filters if filters else Filter()
        request = EvaluationRequest(
            dataset_names=[dataset.name for dataset in datasets],  # type: ignore - issue #604
            model_names=[self.name],  # type: ignore - issue #604
            filters=filters,
            parameters=EvaluationParameters(
                task_type=TaskType.TEXT_GENERATION,
                metrics_to_return=metrics_to_return,
                llm_api_params=llm_api_params,
                bleu_weights=bleu_weights,
                rouge_types=rouge_types,
                rouge_use_stemmer=rouge_use_stemmer,
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def delete(self, timeout: int = 0):
        """
        Delete the `Model` object from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_model(self.name, timeout)  # type: ignore

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given model.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the model.
        """
        return Client(self.conn).get_labels_from_model(self)

    def get_evaluations(
        self,
        metrics_to_sort_by: Optional[
            Dict[str, Union[Dict[str, str], str]]
        ] = None,
        *_,
        timeout: Optional[float] = None,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given model.

        Parameters
        ----------
        metrics_to_sort_by : dict[str, str | dict[str, str]], optional
            An optional dict of metric types to sort the evaluations by.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the model.
        """
        return Client(self.conn).get_evaluations(
            models=[self],
            metrics_to_sort_by=metrics_to_sort_by,
            timeout=timeout,
        )


class Client:
    """
    Valor client object for interacting with the api.

    Parameters
    ----------
    connection : ClientConnection, optional
        Option to use an existing connection object.
    """

    def __init__(self, connection: Optional[ClientConnection] = None):
        if not connection:
            connection = get_connection()
        self.conn = connection

    @classmethod
    def connect(
        cls,
        host: str,
        access_token: Optional[str] = None,
        reconnect: bool = False,
    ) -> Client:
        """
        Establishes a connection to the Valor API.

        Parameters
        ----------
        host : str
            The host to connect to. Should start with "http://" or "https://".
        access_token : str
            The access token for the host (if the host requires authentication).
        """
        connect(host=host, access_token=access_token, reconnect=reconnect)
        return cls(get_connection())

    def get_labels(
        self,
        filters: Optional[Filter] = None,
    ) -> List[Label]:
        """
        Gets all labels using an optional filter.

        Parameters
        ----------
        filters : Filter, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        filters = filters if filters is not None else Filter()
        return [
            Label(**label) for label in self.conn.get_labels(filters.to_dict())
        ]

    def get_labels_from_dataset(
        self, dataset: Union[Dataset, str]
    ) -> List[Label]:
        """
        Get all labels associated with a dataset's ground truths.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        return [
            Label(**label)
            for label in self.conn.get_labels_from_dataset(dataset_name)  # type: ignore
        ]

    def get_labels_from_model(self, model: Union[Model, str]) -> List[Label]:
        """
        Get all labels associated with a model's ground truths.

        Parameters
        ----------
        model : valor.Model
            The model to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        model_name = model.name if isinstance(model, Model) else model
        return [
            Label(**label)
            for label in self.conn.get_labels_from_model(model_name)  # type: ignore
        ]

    def create_dataset(
        self,
        dataset: Union[Dataset, dict],
    ) -> None:
        """
        Creates a dataset.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to create.
        """
        if isinstance(dataset, Dataset):
            dataset = dataset.encode_value()
        self.conn.create_dataset(dataset)

    def create_groundtruths(
        self,
        dataset: Dataset,
        groundtruths: List[GroundTruth],
        ignore_existing_datums: bool = False,
        timeout: Optional[float] = None,
    ):
        """
        Creates ground truths.

        Parameters
        ----------

        dataset : valor.Dataset
            The dataset to create the ground truth for.
        groundtruths : List[valor.GroundTruth]
            The ground truths to create.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.
        ignore_existing_datums : bool, default=False
            If True, will ignore datums that already exist in the backend.
            If False, will raise an error if any datums already exist.
            Default is False.
        """
        groundtruths_json = []
        for groundtruth in groundtruths:
            if not isinstance(groundtruth, GroundTruth):
                raise TypeError(
                    f"Expected ground truth to be of type 'valor.GroundTruth' not '{type(groundtruth)}'."
                )
            if not isinstance(groundtruth.annotations._value, list):
                raise TypeError
            groundtruth_dict = groundtruth.encode_value()
            groundtruth_dict["dataset_name"] = dataset.name
            groundtruths_json.append(groundtruth_dict)
        self.conn.create_groundtruths(
            groundtruths_json,
            timeout=timeout,
            ignore_existing_datums=ignore_existing_datums,
        )

    def get_groundtruth(
        self,
        dataset: Union[Dataset, str],
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        dataset : Union[Dataset, str]
            The dataset the datum belongs to.
        datum : Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        datum_uid = datum.uid if isinstance(datum, Datum) else datum
        try:
            resp = self.conn.get_groundtruth(
                dataset_name=dataset_name, datum_uid=datum_uid  # type: ignore
            )
            resp.pop("dataset_name")
            return GroundTruth.decode_value(resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def finalize_dataset(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes a dataset such that new ground truths cannot be added to it.

        Parameters
        ----------
        dataset : str
            The dataset to be finalized.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        return self.conn.finalize_dataset(name=dataset_name)  # type: ignore

    def get_dataset(
        self,
        name: str,
    ) -> Union[Dataset, None]:
        """
        Gets a dataset by name.

        Parameters
        ----------
        name : str
            The name of the dataset to fetch.

        Returns
        -------
        Union[Dataset, None]
            A Dataset with a matching name, or 'None' if one doesn't exist.
        """
        dataset = Dataset.decode_value(
            {
                **self.conn.get_dataset(name),
                "connection": self.conn,
            }
        )
        return dataset

    def get_datasets(
        self,
        filters: Optional[Filter] = None,
    ) -> List[Dataset]:
        """
        Get all datasets, with an option to filter results according to some user-defined parameters.

        Parameters
        ----------
        filters : Filter, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Dataset]
            A list of datasets.
        """
        dataset_list = []
        filters = filters if filters is not None else Filter()
        for kwargs in self.conn.get_datasets(filters.to_dict()):
            dataset = Dataset.decode_value({**kwargs, "connection": self.conn})
            dataset_list.append(dataset)
        return dataset_list

    def get_datums(
        self,
        filters: Optional[Filter] = None,
    ) -> List[Datum]:
        """
        Get all datums using an optional filter.

        Parameters
        ----------
        filters : Filter, optional
            Optional constraints to filter by.

        Returns
        -------
        List[valor.Datum]
            A list datums.
        """

        filters = filters if filters is not None else Filter()
        return [
            Datum.decode_value(datum)
            for datum in self.conn.get_datums(filters.to_dict())
        ]

    def get_datum(
        self,
        dataset: Union[Dataset, str],
        uid: str,
    ) -> Union[Datum, None]:
        """
        Get datum.
        `GET` endpoint.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset the datum belongs to.
        uid : str
            The UID of the datum.

        Returns
        -------
        valor.Datum
            The requested datum or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        resp = self.conn.get_datum(dataset_name=dataset_name, uid=uid)  # type: ignore
        return Datum.decode_value(resp)

    def get_dataset_status(
        self,
        name: str,
    ) -> Union[TableStatus, None]:
        """
        Get the state of a given dataset.

        Parameters
        ----------
        name : str
            The name of the dataset we want to fetch the state of.

        Returns
        ------
        TableStatus | None
            The state of the dataset, or 'None' if the dataset does not exist.
        """
        try:
            return self.conn.get_dataset_status(name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_dataset_summary(
        self,
        name: str,
        *_,
        timeout: Optional[float] = None,
    ) -> DatasetSummary:
        """
        Gets the summary of a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to create a summary for.

        Returns
        -------
        DatasetSummary
            A dataclass containing the dataset summary.
        """
        return DatasetSummary(
            **self.conn.get_dataset_summary(name, timeout=timeout)
        )

    def delete_dataset(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the dataset was deleted.
        """
        self.conn.delete_dataset(name)
        if timeout:
            for _ in range(timeout):
                try:
                    self.get_dataset(name)
                except DatasetDoesNotExistError:
                    break
                time.sleep(1)
            else:
                raise TimeoutError(
                    "Dataset wasn't deleted within timeout interval"
                )

    def create_model(
        self,
        model: Union[Model, dict],
    ):
        """
        Creates a model.

        Parameters
        ----------
        model : valor.Model
            The model to create.
        """
        if isinstance(model, Model):
            model = model.encode_value()
        self.conn.create_model(model)

    def create_predictions(
        self,
        dataset: Dataset,
        model: Model,
        predictions: List[Prediction],
        timeout: Optional[float] = None,
    ) -> None:
        """
        Creates predictions.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        model : valor.Model
            The model making the prediction.
        predictions : List[valor.Prediction]
            The predictions to create.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.
        """
        predictions_json = []
        for prediction in predictions:
            if not isinstance(prediction, Prediction):
                raise TypeError(
                    f"Expected prediction to be of type 'valor.Prediction' not '{type(prediction)}'."
                )
            if not isinstance(prediction.annotations._value, list):
                raise TypeError
            prediction_dict = prediction.encode_value()
            prediction_dict["dataset_name"] = dataset.name
            prediction_dict["model_name"] = model.name
            predictions_json.append(prediction_dict)
        self.conn.create_predictions(predictions_json, timeout=timeout)

    def get_prediction(
        self,
        dataset: Union[Dataset, str],
        model: Union[Model, str],
        datum: Union[Datum, str],
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset : Union[Dataset, str]
            The dataset the datum belongs to.
        model : Union[Model, str]
            The model that made the prediction.
        datum : Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.name if isinstance(model, Model) else model
        datum_uid = datum.uid if isinstance(datum, Datum) else datum

        resp = self.conn.get_prediction(
            dataset_name=dataset_name,  # type: ignore
            model_name=model_name,  # type: ignore
            datum_uid=datum_uid,  # type: ignore
        )
        resp.pop("dataset_name")
        resp.pop("model_name")
        return Prediction.decode_value(resp)

    def finalize_inferences(
        self, dataset: Union[Dataset, str], model: Union[Model, str]
    ) -> None:
        """
        Finalizes a model-dataset pairing such that new predictions cannot be added to it.
        """
        dataset_name = (
            dataset.name if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.name if isinstance(model, Model) else model
        return self.conn.finalize_inferences(
            dataset_name=dataset_name,  # type: ignore
            model_name=model_name,  # type: ignore
        )

    def get_model(
        self,
        name: str,
    ) -> Union[Model, None]:
        """
        Gets a model by name.

        Parameters
        ----------
        name : str
            The name of the model to fetch.

        Returns
        -------
        Union[valor.Model, None]
            A Model with matching name or 'None' if one doesn't exist.
        """
        return Model.decode_value(
            {
                **self.conn.get_model(name),
                "connection": self.conn,
            }
        )

    def get_models(
        self,
        filters: Optional[Filter] = None,
    ) -> List[Model]:
        """
        Get all models using an optional filter.

        Parameters
        ----------
        filters : Filter, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Model]
            A list of models.
        """
        model_list = []
        filters = filters if filters is not None else Filter()
        for kwargs in self.conn.get_models(filters.to_dict()):
            model = Model.decode_value({**kwargs, "connection": self.conn})
            model_list.append(model)
        return model_list

    def get_model_status(
        self,
        dataset_name: str,
        model_name: str,
    ) -> Optional[TableStatus]:
        """
        Get the state of a given model over a dataset.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset that the model is operating over.
        model_name : str
            The name of the model we want to fetch the state of.

        Returns
        ------
        Union[TableStatus, None]
            The state of the model or 'None' if the model doesn't exist.
        """
        try:
            return self.conn.get_model_status(dataset_name, model_name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_model_eval_requests(
        self, model: Union[Model, str]
    ) -> List[Evaluation]:
        """
        Get all evaluations that have been created for a model.

        This does not return evaluation results.

        `GET` endpoint.

        Parameters
        ----------
        model : str
            The model to search by.

        Returns
        -------
        List[Evaluation]
            A list of evaluations.
        """
        model_name = model.name if isinstance(model, Model) else model
        return [
            Evaluation(**evaluation, connection=self.conn)
            for evaluation in self.conn.get_model_eval_requests(model_name)  # type: ignore
        ]

    def delete_model(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a model.

        Parameters
        ----------
        name : str
            The name of the model to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the model was deleted.
        """
        self.conn.delete_model(name)
        if timeout:
            for _ in range(timeout):
                try:
                    self.get_model(name)
                except ModelDoesNotExistError:
                    break
                time.sleep(1)
            else:
                raise TimeoutError(
                    "Model wasn't deleted within timeout interval"
                )

    def get_evaluations(
        self,
        *,
        evaluation_ids: Optional[List[int]] = None,
        models: Union[List[Model], List[str], None] = None,
        datasets: Union[List[Dataset], List[str], None] = None,
        metrics_to_sort_by: Optional[
            Dict[str, Union[Dict[str, str], str]]
        ] = None,
        timeout: Optional[float] = None,
    ) -> List[Evaluation]:
        """
        Returns all evaluations associated with user-supplied dataset and/or model names.

        Parameters
        ----------
        evaluation_ids : List[int], optional.
            A list of job IDs to return metrics for.
        models : Union[List[valor.Model], List[str]], optional
            A list of model names that we want to return metrics for.
        datasets : Union[List[valor.Dataset], List[str]], optional
            A list of dataset names that we want to return metrics for.
        metrics_to_sort_by : dict[str, str | dict[str, str]], optional
            An optional dict of metric types to sort the evaluations by.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        List[valor.Evaluation]
            A list of evaluations.
        """
        if isinstance(datasets, list):
            datasets = [  # type: ignore
                element.name if isinstance(element, Dataset) else element
                for element in datasets
            ]
        if isinstance(models, list):
            models = [  # type: ignore
                element.name if isinstance(element, Model) else element
                for element in models
            ]
        return [
            Evaluation(connection=self.conn, **evaluation)
            for evaluation in self.conn.get_evaluations(
                evaluation_ids=evaluation_ids,
                models=models,  # type: ignore
                datasets=datasets,  # type: ignore
                metrics_to_sort_by=metrics_to_sort_by,
                timeout=timeout,
            )
        ]

    def evaluate(
        self,
        request: EvaluationRequest,
        *_,
        allow_retries: bool = False,
        timeout: Optional[float] = None,
    ) -> List[Evaluation]:
        """
        Creates as many evaluations as necessary to fulfill the request.

        Parameters
        ----------
        request : schemas.EvaluationRequest
            The requested evaluation parameters.
        allow_retries : bool, default = False
            Option to retry previously failed evaluations.
        timeout : float, optional
            The number of seconds the client should wait until raising a timeout.

        Returns
        -------
        List[Evaluation]
            A list of evaluations that meet the parameters.
        """
        return [
            Evaluation(**evaluation)
            for evaluation in self.conn.evaluate(
                request.to_dict(),
                allow_retries=allow_retries,
                timeout=timeout,
            )
        ]

    def delete_evaluation(self, evaluation_id: int, timeout: int = 0) -> None:
        """
        Deletes an evaluation.

        Parameters
        ----------
        evaluation_id : int
            The id of the evaluation to be deleted.
        timeout : int, default=0
            The number of seconds to wait in order to confirm that the model was deleted.
        """
        self.conn.delete_evaluation(evaluation_id)
        if timeout:
            for _ in range(timeout):
                try:
                    self.get_evaluations(evaluation_ids=[evaluation_id])
                except EvaluationDoesNotExist:
                    break
                time.sleep(1)
            else:
                raise TimeoutError(
                    "Evaluation wasn't deleted within timeout interval"
                )



from dataclasses import asdict, dataclass
from typing import Optional, Union

from valor.schemas.symbolic.operators import (
    And,
    Contains,
    Eq,
    FunctionType,
    Gt,
    Gte,
    Inside,
    Intersects,
    IsNotNull,
    IsNull,
    Lt,
    Lte,
    Ne,
    Not,
    Or,
    Outside,
)

FunctionTypeTuple = (
    And,
    Or,
    Not,
    IsNull,
    IsNotNull,
    Eq,
    Ne,
    Gt,
    Gte,
    Lt,
    Lte,
    Intersects,
    Inside,
    Outside,
    Contains,
)


@dataclass
class Filter:
    """
    A data class that encapsulates filter conditions for various Valor components.

    Attributes
    ----------
    datasets : dict | FunctionType, optional
        Filter conditions to apply to datasets.
    models : dict | FunctionType, optional
        Filter conditions to apply to models.
    datums : dict | FunctionType, optional
        Filter conditions to apply to datums.
    annotations : dict | FunctionType, optional
        Filter conditions to apply to annotations.
    groundtruths : dict | FunctionType, optional
        Filter conditions to apply to groundtruths.
    predictions : dict | FunctionType, optional
        Filter conditions to apply to predictions.
    labels : dict | FunctionType, optional
        Filter conditions to apply to labels.
    embeddings : dict | FunctionType, optional
        Filter conditions to apply to embeddings.

    Examples
    --------
    Filter annotations by area and label.
    >>> Filter(
    ...     annotations=And(
    ...         Label.key == "name",
    ...         Annotation.raster.area > upper_bound,
    ...     )
    ... )

    Filter datums by annotations and labels.
    >>> Filter(
    ...     datums=And(
    ...         Label.key == "name",
    ...         Annotation.raster.area > upper_bound,
    ...     )
    ... )
    """

    datasets: Optional[Union[dict, FunctionType]] = None
    models: Optional[Union[dict, FunctionType]] = None
    datums: Optional[Union[dict, FunctionType]] = None
    annotations: Optional[Union[dict, FunctionType]] = None
    groundtruths: Optional[Union[dict, FunctionType]] = None
    predictions: Optional[Union[dict, FunctionType]] = None
    labels: Optional[Union[dict, FunctionType]] = None
    embeddings: Optional[Union[dict, FunctionType]] = None

    def to_dict(self) -> dict:
        if isinstance(self.datasets, FunctionTypeTuple):
            self.datasets = self.datasets.to_dict()
        if isinstance(self.models, FunctionTypeTuple):
            self.models = self.models.to_dict()
        if isinstance(self.datums, FunctionTypeTuple):
            self.datums = self.datums.to_dict()
        if isinstance(self.annotations, FunctionTypeTuple):
            self.annotations = self.annotations.to_dict()
        if isinstance(self.groundtruths, FunctionTypeTuple):
            self.groundtruths = self.groundtruths.to_dict()
        if isinstance(self.predictions, FunctionTypeTuple):
            self.predictions = self.predictions.to_dict()
        if isinstance(self.labels, FunctionTypeTuple):
            self.labels = self.labels.to_dict()
        if isinstance(self.embeddings, FunctionTypeTuple):
            self.embeddings = self.embeddings.to_dict()
        return asdict(self)



import numpy as np
import PIL.Image
import pytest

from valor import Annotation, GroundTruth, Label
from valor.metatypes import ImageMetadata
from valor.schemas import Box, MultiPolygon, Polygon, Raster
from valor.viz import (
    _draw_detection_on_image,
    _polygons_to_binary_mask,
    create_combined_segmentation_mask,
    draw_bounding_box_on_image,
    draw_detections_on_image,
    draw_raster_on_image,
)


@pytest.fixture
def bounding_box() -> Box:
    return Box(
        value=[
            [
                (107, 207),
                (107, 307),
                (207, 307),
                (207, 207),
                (107, 207),
            ]
        ]
    )


@pytest.fixture
def bounding_poly() -> Polygon:
    return Polygon(
        [
            [
                (100, 100),
                (200, 100),
                (200, 200),
                (100, 200),
                (100, 100),
            ]
        ]
    )


@pytest.fixture
def poly1(bounding_poly: Polygon) -> Polygon:
    return Polygon(
        [
            bounding_poly.get_value()[0],
            [
                (150, 120),
                (180, 120),
                (180, 140),
                (150, 140),
                (150, 120),
            ],
        ]
    )


def test__polygons_to_binary_mask(poly1):
    poly2 = Polygon(
        [
            [
                (10, 15),
                (20, 15),
                (20, 20),
                (10, 20),
                (10, 15),
            ]
        ]
    )

    mask = _polygons_to_binary_mask([poly1, poly2], 500, 600)

    area_poly1 = (200 - 100 + 1) * (200 - 100 + 1) - (180 - 150 + 1) * (
        140 - 120 + 1
    )

    area_poly2 = (20 - 10 + 1) * (20 - 15 + 1)

    assert mask.sum() == area_poly1 + area_poly2


def test_create_combined_segmentation_mask(poly1: Polygon):
    image = ImageMetadata.create(uid="uid", height=200, width=200).datum

    gt1 = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v2"),
                    Label(key="k3", value="v3"),
                ],
                raster=Raster.from_geometry(
                    MultiPolygon([poly1.get_value()]),
                    height=2,
                    width=2,
                ),
            ),
            Annotation(
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v3"),
                ],
                raster=Raster.from_numpy(
                    np.array([[True, False], [False, True]]),
                ),
            ),
        ],
    )

    gt_with_size_mismatch = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v2"),
                    Label(key="k3", value="v3"),
                ],
                raster=Raster.from_geometry(
                    MultiPolygon([poly1.get_value()]),
                    height=20,
                    width=20,
                ),
            ),
            Annotation(
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v3"),
                ],
                raster=Raster.from_numpy(
                    np.array([[True, False], [False, True]]),
                ),
            ),
        ],
    )

    # test that a size mistmatch between rasters is caught
    with pytest.raises(ValueError) as exc_info:
        create_combined_segmentation_mask(
            gt_with_size_mismatch,
            label_key="k2",
            filter_on_instance_segmentations=False,
        )
    assert "(20, 20) != (2, 2)" in str(exc_info)

    # check get an error since "k3" isn't a label key in seg2
    with pytest.raises(RuntimeError) as exc_info:
        create_combined_segmentation_mask(
            gt1,
            label_key="k3",
            filter_on_instance_segmentations=True,
        )
    assert "doesn't have a label" in str(exc_info)

    # should have one distinct (non-black) color
    combined_mask, legend = create_combined_segmentation_mask(
        gt1,
        label_key="k1",
        filter_on_instance_segmentations=False,
    )
    combined_mask = np.array(combined_mask)
    # check that we get two unique RGB values (black and one color for label value "v1")
    assert combined_mask.shape == (2, 2, 3)
    assert len(legend) == 1  # background color 'black' is not counted

    # should have two distinct (non-black) color
    combined_mask, legend = create_combined_segmentation_mask(
        gt1,
        label_key="k2",
        filter_on_instance_segmentations=False,
    )
    combined_mask = np.array(combined_mask)
    # check that we get three unique RGB values (black and two colors for label values "v2" and "v3")
    assert combined_mask.shape == (2, 2, 3)
    assert len(legend) == 2  # background color 'black' is not counted


def test__draw_detections_on_image(bounding_poly: Polygon, bounding_box: Box):
    # test polygon
    poly_detection = Annotation(
        labels=[Label(key="k", value="v")],
        polygon=bounding_poly,
    )

    img = PIL.Image.new("RGB", (300, 300))

    output = _draw_detection_on_image(
        detection=poly_detection, img=img, inplace=True
    )

    assert output.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(output).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0

    # test bounding box
    poly_detection = Annotation(
        labels=[Label(key="k", value="v")], bounding_box=bounding_box
    )

    img = PIL.Image.new("RGB", (300, 300))

    output = _draw_detection_on_image(
        detection=poly_detection, img=img, inplace=True
    )

    assert output.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(output).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0


def test_draw_detections_on_image(bounding_poly: Polygon):
    detections = [
        GroundTruth(
            datum=ImageMetadata.create("test", 300, 300).datum,
            annotations=[
                Annotation(
                    labels=[Label(key="k", value="v")],
                    polygon=bounding_poly,
                )
            ],
        ),
    ]
    img = PIL.Image.new("RGB", (300, 300))

    output = draw_detections_on_image(detections, img)

    assert output.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(output).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0


def test_draw_raster_on_image(raster):

    img = PIL.Image.new("RGB", (20, 20))

    output = draw_raster_on_image(raster, img)

    assert output.size == (20, 20)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(output).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0

    # test errors
    img2 = PIL.Image.new("RGB", (300, 300))
    with pytest.raises(ValueError):
        draw_raster_on_image(raster, img2)


def test_draw_bounding_box_on_image():
    box = Box(
        value=[
            [
                (10, 20),
                (10, 30),
                (20, 30),
                (20, 20),
                (10, 20),
            ]
        ]
    )

    img = PIL.Image.new("RGB", (300, 300))

    img = draw_bounding_box_on_image(bounding_box=box, img=img)

    assert img.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(img).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0



from unittest.mock import patch

import pytest

from valor.client import (
    _format_request_timeout,
    connect,
    get_connection,
    reset_connection,
)
from valor.exceptions import (
    ClientAlreadyConnectedError,
    ClientConnectionFailed,
    ClientNotConnectedError,
)


def test__format_request_timeout():
    assert _format_request_timeout(timeout=None, default=30) == 30
    assert _format_request_timeout(timeout=60, default=30) == 60
    assert _format_request_timeout(timeout=-1, default=30) is None
    assert _format_request_timeout(timeout=0, default=30) is None
    assert _format_request_timeout(timeout=-0.1, default=30) is None


@patch("valor.client.ClientConnection")
def test_connect(ClientConnection):
    connect(host="host")
    ClientConnection.assert_called_once()

    with pytest.raises(ClientAlreadyConnectedError):
        connect(host="host")

    connect(host="host", reconnect=True)

    ClientConnection.side_effect = ClientConnectionFailed("testing")
    with pytest.raises(ClientConnectionFailed) as e:
        connect(host="host", reconnect=True)
    assert "testing" in str(e)


@patch("valor.client.ClientConnection")
def test_get_connection(ClientConnection):
    reset_connection()

    with pytest.raises(ClientNotConnectedError):
        get_connection()

    connect(host="host")
    ClientConnection.assert_called_once()


@patch("valor.client.ClientConnection")
def test_reset_connection(ClientConnection):
    connect(host="host", reconnect=True)
    assert get_connection() is not None

    reset_connection()

    with pytest.raises(ClientNotConnectedError):
        get_connection()
    connect(host="host")  # test without reconnect arg
    assert get_connection() is not None



import math
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from PIL import Image, ImageDraw, ImageFont

from valor import Annotation, GroundTruth, Prediction, schemas

# https://sashat.me/2017/01/11/list-of-20-simple-distinct-colors/
COLOR_MAP = [
    (230, 25, 75),
    (60, 180, 75),
    (255, 225, 25),
    (0, 130, 200),
    (245, 130, 48),
    (145, 30, 180),
    (70, 240, 240),
    (240, 50, 230),
    (210, 245, 60),
    (250, 190, 190),
    (0, 128, 128),
    (230, 190, 255),
    (170, 110, 40),
    (255, 250, 200),
    (128, 0, 0),
    (170, 255, 195),
    (128, 128, 0),
    (255, 215, 180),
    (0, 0, 128),
    (128, 128, 128),
]


def _polygons_to_binary_mask(
    polys: List[schemas.Polygon], img_w, img_h
) -> np.ndarray:
    """note there's some aliasing/areas differences between this
    and e.g. postgis, so this method should only be used for visualization
    """
    mask = Image.new("1", (img_w, img_h), (False,))
    draw = ImageDraw.Draw(mask)
    for poly in polys:
        draw.polygon(poly.boundary, fill=(True,))  # type: ignore
        if poly.holes is not None:
            for hole in poly.holes:
                draw.polygon(hole, fill=(False,))  # type: ignore

    return np.array(mask)


def create_combined_segmentation_mask(
    annotated_datum: Union[GroundTruth, Prediction],
    label_key: str,
    filter_on_instance_segmentations: bool = False,
) -> Tuple[Image.Image, Dict[str, Image.Image]]:
    """
    Creates a combined segmentation mask from a list of segmentations.

    Parameters
    -------
    annotated_datum : Union[GroundTruth, Prediction]
        A list of segmentations. These all must have the same `image` attribute.
    label_key : str
        The label key to use.
    filter_on_instance_segmentations : bool, optional
        Whether to filter on instance segmentations or not.

    Returns
    -------
    tuple
        The first element of the tuple is the combined mask, as an RGB PIL image. The second
        element is a color legend: it's a dict with the unique labels as keys and the
        PIL image swatches as values.

    Raises
    ------
    RuntimeError
        If all segmentations don't belong to the same image or there is a
        segmentation that doesn't have `label_key` as the key of one of its labels.
    ValueError
        If there aren't any segmentations.
    """

    # validate input type
    if not isinstance(annotated_datum, (GroundTruth, Prediction)):
        raise ValueError("Expected either a 'GroundTruth' or 'Prediction'")

    # verify there are a nonzero number of annotations
    if len(annotated_datum.annotations) == 0:
        raise ValueError("annotations cannot be empty.")

    # validate raster size
    img_h = None
    img_w = None
    for annotation in annotated_datum.annotations:
        raster = annotation.raster
        if raster.get_value() is None:
            raise ValueError("No raster exists.")
        if img_h is None:
            img_h = raster.height
        if img_w is None:
            img_w = raster.width
        if (img_h != raster.height) or (img_w != raster.width):
            raise ValueError(
                f"Size mismatch between rasters. {(img_h, img_w)} != {(raster.height, raster.width)}"
            )
    if img_h is None or img_w is None:
        raise ValueError(
            f"Segmentation bounds not properly defined. {(img_h, img_w)}"
        )

    # unpack raster annotations
    annotations: List[Annotation] = []
    for annotation in annotated_datum.annotations:
        if (
            annotation.is_instance or False
        ) == filter_on_instance_segmentations:
            annotations.append(annotation)

    # unpack label values
    label_values = []
    for annotation in annotations:
        for label in annotation.labels:
            if label.key == label_key:
                label_values.append(label.value)
    if not label_values:
        raise RuntimeError(
            f"Annotation doesn't have a label with key `{label_key}`"
        )

    # assign label coloring
    unique_label_values = list(set(label_values))
    label_value_to_color = {
        v: COLOR_MAP[i] for i, v in enumerate(unique_label_values)
    }
    seg_colors = [label_value_to_color[v] for v in label_values]

    # create mask
    combined_mask = np.zeros((img_h, img_w, 3), dtype=np.uint8)
    for annotation, color in zip(annotations, seg_colors):
        raster = annotation.raster
        if raster.get_value() is None:
            raise ValueError("No raster exists.")
        if raster.array is not None:
            if raster.geometry is None:
                mask = raster.array
            elif isinstance(raster.geometry, schemas.MultiPolygon):
                mask = _polygons_to_binary_mask(
                    raster.geometry.to_polygons(),
                    img_w=img_w,
                    img_h=img_h,
                )
            elif isinstance(raster.geometry, (schemas.Box, schemas.Polygon)):
                mask = _polygons_to_binary_mask(
                    [raster.geometry],
                    img_w=img_w,
                    img_h=img_h,
                )
            else:
                continue
            combined_mask[np.where(mask)] = color
        else:
            continue

    legend = {
        v: Image.new("RGB", (20, 20), color)
        for v, color in label_value_to_color.items()
    }

    return Image.fromarray(combined_mask), legend


def draw_bounding_box_on_image(
    bounding_box: schemas.Box,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
) -> Image.Image:
    """Draws a bounding polygon on an image. This operation is not done in place.

    Parameters
    ----------
    bounding_box
        Bounding box to draw on the image.
    img
        Pillow image to draw on.
    color
        RGB tuple of the color to use.

    Returns
    -------
    img
        Pillow image with bounding box drawn on it.
    """
    coords = bounding_box.get_value()
    return _draw_bounding_polygon_on_image(
        schemas.Polygon(coords), img, color=color, inplace=False
    )


def _draw_detection_on_image(
    detection: Annotation, img: Image.Image, inplace: bool
) -> Image.Image:
    """Draw a detection on an image."""
    text = ", ".join(
        [f"{label.key}:{label.value}" for label in detection.labels]
    )
    box = detection.bounding_box
    polygon = detection.polygon
    if polygon is not None:
        img = _draw_bounding_polygon_on_image(
            polygon,
            img,
            inplace=inplace,
            text=text,
        )
    elif box.get_value() is not None:
        img = _draw_bounding_polygon_on_image(
            box,
            img,
            inplace=inplace,
            text=text,
        )

    return img


def _draw_bounding_polygon_on_image(
    polygon: schemas.Polygon,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
    inplace: bool = False,
    text: Optional[str] = None,
    font_size: int = 24,
) -> Image.Image:
    """Draw a bounding polygon on an image."""
    img = img if inplace else img.copy()
    img_draw = ImageDraw.Draw(img)

    img_draw.polygon(
        [p for p in polygon.boundary],
        outline=color,
    )

    if text is not None:
        _write_text(
            font_size=font_size,
            text=text,
            boundary=polygon,
            draw=img_draw,
            color=color,
        )
    return img


def _write_text(
    font_size: int,
    text: str,
    boundary: schemas.Polygon,
    draw: ImageDraw.ImageDraw,
    color: Union[Tuple[int, int, int], str],
):
    """Write text on an image."""
    try:
        font = ImageFont.truetype("arial.ttf", font_size)
    except IOError:
        try:
            font = ImageFont.truetype("Arial.ttf", font_size)
        except IOError:
            font = ImageFont.load_default()

    _, _, text_width, text_height = font.getbbox(text)
    if boundary.ymin > text_height:
        text_bottom = boundary.ymin
    else:
        text_bottom = boundary.ymax + text_height

    margin = math.ceil(0.05 * text_height) - 1
    draw.rectangle(
        (
            (boundary.xmin, text_bottom - text_height - 2 * margin),
            (boundary.xmin + text_width, text_bottom),
        ),
        fill=color,
    )
    draw.text(
        (boundary.xmin + margin, text_bottom - text_height - margin),
        text,
        fill="black",
        font=font,
    )


def draw_raster_on_image(
    raster: schemas.Raster,
    img: Image.Image,
    color: Tuple[int, int, int] = (255, 0, 0),
    alpha: float = 0.4,
) -> Image.Image:
    """Draws the raster on top of an image. This operation is not done in place.

    Parameters
    ----------
    img
        pillow image to draw on.
    color
        RGB tuple of the color to use
    alpha
        alpha (transparency) value of the mask. 0 is fully transparent, 1 is fully opaque
    """
    img = img.copy()
    binary_mask = raster.array
    mask_arr = np.zeros(
        (binary_mask.shape[0], binary_mask.shape[1], 3), dtype=np.uint8
    )
    mask_arr[binary_mask] = color
    mask_img = Image.fromarray(mask_arr)

    if mask_img.size != img.size:
        raise ValueError("Input image and raster must be the same size.")
    blend = Image.blend(img, mask_img, alpha=alpha)
    img.paste(blend, (0, 0), mask=Image.fromarray(binary_mask))

    return img


def draw_detections_on_image(
    detections: Sequence[Union[GroundTruth, Prediction]],
    img: Image.Image,
) -> Image.Image:
    """
    Draws detections (bounding boxes and labels) on an image.
    Parameters
    -------
    detections : List[Union[GroundTruth, Prediction]]
        A list of `GroundTruths` or `Predictions` to draw on the image.
    img : Image.Image
        The image to draw the detections on.
    Returns
    -------
    img : Image.Image
        An image with the detections drawn on.
    """

    annotations = []
    for datum in detections:
        annotations.extend(datum.annotations)

    for i, detection in enumerate(annotations):
        if detection.raster and detection.is_instance is True:
            img = _draw_detection_on_image(detection, img, inplace=i != 0)
    return img



import datetime

import pytest

from valor.schemas import (
    Boolean,
    Date,
    DateTime,
    Duration,
    Float,
    Integer,
    LineString,
    List,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    String,
    Symbol,
    Time,
    Variable,
)
from valor.schemas.symbolic.operators import Condition, Eq, Function, Ne
from valor.schemas.symbolic.types import (
    Dictionary,
    DictionaryValue,
    _get_type_by_value,
    get_type_by_name,
)


def test__get_type_by_value():
    assert _get_type_by_value(True) is Boolean
    assert _get_type_by_value("hello world") is String
    assert _get_type_by_value(int(1)) is Integer
    assert _get_type_by_value(float(3.14)) is Float
    assert (
        _get_type_by_value(datetime.datetime(year=2024, month=1, day=1))
        is DateTime
    )
    assert _get_type_by_value(datetime.date(year=2024, month=1, day=1)) is Date
    assert (
        _get_type_by_value(datetime.time(hour=1, minute=1, second=1)) is Time
    )
    assert _get_type_by_value(datetime.timedelta(seconds=100)) is Duration
    assert _get_type_by_value((1, 1)) is Point
    assert _get_type_by_value([(1, 1)]) is MultiPoint
    assert _get_type_by_value([(1, 1), (2, 2)]) is LineString
    assert _get_type_by_value([[(1, 1), (2, 2)]]) is MultiLineString
    assert _get_type_by_value([[(1, 1), (2, 2), (0, 1), (1, 1)]]) is Polygon
    assert (
        _get_type_by_value([[[(1, 1), (2, 2), (0, 1), (1, 1)]]])
        is MultiPolygon
    )
    assert _get_type_by_value({"randomvalue": "idk"}) is Dictionary
    with pytest.raises(NotImplementedError):
        assert _get_type_by_value(set()).__name__


def test_get_type_by_name():
    types_ = [
        Boolean,
        String,
        Integer,
        Float,
        DateTime,
        Date,
        Time,
        Duration,
        Point,
        MultiPoint,
        LineString,
        MultiLineString,
        Polygon,
        MultiPolygon,
    ]
    for type_ in types_:
        type_name = type_.__name__
        assert issubclass(type_, Variable)
        assert isinstance(type_name, str)
        assert get_type_by_name(type_name) is type_
        assert get_type_by_name(f"list[{type_name}]") is List[type_]
    with pytest.raises(NotImplementedError):
        assert get_type_by_name("some_nonexistent_type")


def get_function_name(fn: str) -> str:
    fns = {
        "__eq__": "eq",
        "__ne__": "ne",
        "__and__": "and",
        "__or__": "or",
        "__xor__": "xor",
        "__gt__": "gt",
        "__ge__": "ge",
        "__lt__": "lt",
        "__le__": "le",
        "is_none": "isnull",
        "is_not_none": "isnotnull",
        "intersects": "intersects",
        "inside": "inside",
        "outside": "outside",
    }
    return fns[fn]


def _test_to_dict(objcls, value):
    # test __init__
    assert objcls(value).to_dict() == {
        "type": objcls.__name__.lower(),
        "value": objcls(value).encode_value(),
    }
    # test value
    assert objcls(value).to_dict() == {
        "type": objcls.__name__.lower(),
        "value": objcls(value).encode_value(),
    }
    # test symbolic
    assert objcls.symbolic().to_dict() == {
        "name": objcls.__name__.lower(),
        "key": None,
    }


def _test_generic(objcls, permutations, op):
    """Tests expressions that can only be resolved to JSON."""
    for a, _ in permutations:
        A = objcls(a)
        C = objcls.symbolic()
        # test variable -> builtin against variable -> variable
        assert (
            C.__getattribute__(op)(a).to_dict()
            == C.__getattribute__(op)(A).to_dict()
        )
        # test commutative propery (this will fail)
        with pytest.raises(AssertionError):
            try:
                # function does not exist in left-operand
                a.__getattribute__(op)(C)
                # function exists, but is not commutative
                if type(a.__getattribute__(op)(A)) not in {objcls, type(a)}:
                    raise AssertionError("NotImplementedType")
            except AttributeError as e:
                raise AssertionError(e)
        # test instance dictionary generation
        _test_to_dict(objcls, a)
        # test functional dictionary generation
        expr = C.__getattribute__(op)(a)
        expr_dict = expr.to_dict()
        if isinstance(expr, Ne):
            # this is an edge case as the Ne operator is currently set to Not(Equal(A, B))
            assert len(expr_dict) == 2
            assert expr_dict["op"] == "not"
            assert expr_dict["args"] == Eq(C, A).to_dict()
        elif issubclass(type(expr), Function):
            assert len(expr_dict) == 2
            assert expr_dict["op"] == get_function_name(op)
            assert expr_dict["args"] == [
                C.to_dict(),
                A.to_dict(),
            ]
        elif issubclass(type(expr), Condition):
            assert len(expr_dict) == 3
            assert expr_dict["op"] == get_function_name(op)
            assert expr_dict["lhs"] == C.to_dict()
            assert expr_dict["rhs"] == A.to_dict()
        else:
            raise AssertionError


def _test_unsupported(objcls, permutations, op):
    for a, b in permutations:
        with pytest.raises(AttributeError):
            objcls(a).__getattribute__(op)(b)


def test_list():
    # interoperable with built-in 'list'

    assert isinstance(List[Float], type)
    assert issubclass(List[Float], Variable)

    # test creating symbolic lists
    symbol = List[Float].symbolic()
    assert symbol.__str__() == "list[float]"
    assert symbol.to_dict() == {
        "name": "list[float]",
        "key": None,
    }

    # test creating valued lists
    variable = List[Float]([0.1, 0.2, 0.3])
    assert variable.__str__() == "[Float(0.1), Float(0.2), Float(0.3)]"
    assert variable.to_dict() == {
        "type": "list[float]",
        "value": [0.1, 0.2, 0.3],
    }

    # test setting value in list by index
    assert variable[1].get_value() == 0.2
    variable[1] = 3.14
    assert variable[1].get_value() == 3.14
    variable[1] = Float(0.2)

    # test nested typing
    assert variable[0].get_value() == 0.1

    # test comparison symbol -> value
    eq = symbol == [0.1, 0.2, 0.3]
    assert isinstance(eq, Eq)
    assert eq.to_dict() == {
        "op": "eq",
        "lhs": {
            "name": "list[float]",
            "key": None,
        },
        "rhs": {"type": "list[float]", "value": [0.1, 0.2, 0.3]},
    }

    # test comparison symbol -> valued variable
    eq = symbol == variable
    assert isinstance(eq, Eq)
    assert eq.to_dict() == {
        "op": "eq",
        "lhs": {
            "name": "list[float]",
            "key": None,
        },
        "rhs": {"type": "list[float]", "value": [0.1, 0.2, 0.3]},
    }

    # test decode from json dict
    assert List[Float].decode_value(
        [0.1, 0.2, 0.3]
    ).get_value() == [  # type: ignore - issue #604
        0.1,
        0.2,
        0.3,
    ]

    # test comparison between valued variable and value
    assert variable == [0.1, 0.2, 0.3]

    # test setting list to non-list type
    with pytest.raises(TypeError):
        assert List[String](String("hello"))

    # test setting list item to unsupported type
    with pytest.raises(TypeError):
        assert List[Integer]([String("hello")])

    # test that untyped wrapper is not implemented
    with pytest.raises(TypeError):
        List()  # type: ignore - testing


def test_dictionary_value():
    # test cannot hold a value
    with pytest.raises(ValueError):
        DictionaryValue(1)  # type: ignore - testing

    # test symbol must have key
    with pytest.raises(ValueError) as e:
        DictionaryValue(
            symbol=Symbol(name="a"),
        )
    assert "key" in str(e)

    # test router
    assert (DictionaryValue.symbolic(name="a", key="b") == 0).to_dict()[
        "op"
    ] == "eq"
    assert (DictionaryValue.symbolic(name="a", key="b") != 0).to_dict()[
        "op"
    ] == "not"
    assert (DictionaryValue.symbolic(name="a", key="b") != 0).to_dict()[
        "args"
    ]["op"] == "eq"
    assert (DictionaryValue.symbolic(name="a", key="b") >= 0).to_dict()[
        "op"
    ] == "gte"
    assert (DictionaryValue.symbolic(name="a", key="b") <= 0).to_dict()[
        "op"
    ] == "lte"
    assert (DictionaryValue.symbolic(name="a", key="b") > 0).to_dict()[
        "op"
    ] == "gt"
    assert (DictionaryValue.symbolic(name="a", key="b") < 0).to_dict()[
        "op"
    ] == "lt"
    assert (
        DictionaryValue.symbolic(name="a", key="b").intersects((0, 0))
    ).to_dict()["op"] == "intersects"
    assert (
        DictionaryValue.symbolic(name="a", key="b").inside((0, 0))
    ).to_dict()["op"] == "inside"
    assert (
        DictionaryValue.symbolic(name="a", key="b").outside((0, 0))
    ).to_dict()["op"] == "outside"
    assert (DictionaryValue.symbolic(name="a", key="b").is_none()).to_dict()[
        "op"
    ] == "isnull"
    assert (
        DictionaryValue.symbolic(name="a", key="b").is_not_none()
    ).to_dict()["op"] == "isnotnull"
    eq = DictionaryValue.symbolic(name="a", key="b") == 0
    assert isinstance(eq, Eq)
    assert eq.to_dict()["op"] == "eq"

    # test router with Variable type
    assert (DictionaryValue.symbolic(name="a", key="b") == Float(0)).to_dict()[
        "op"
    ] == "eq"


def test_dictionary():
    # interoperable with built-in 'dict'
    x = {
        "k0": True,
        "k1": "v1",
        "k2": 123,
        "k3": 1.24,
        "k4": datetime.datetime(year=2024, month=1, day=1),
        "k5": datetime.date(year=2024, month=1, day=1),
        "k6": datetime.time(hour=1),
        "k7": datetime.timedelta(seconds=100),
        "k8": Point((1, -1)),
        "k9": MultiPoint([(0, 0), (1, 1)]),
        "k10": LineString([(0, 0), (1, 1)]),
        "k11": MultiLineString([[(0, 0), (1, 1)]]),
        "k12": Polygon([[(0, 0), (1, 1), (0, 1), (0, 0)]]),
        "k13": MultiPolygon([[[(0, 0), (1, 1), (0, 1), (0, 0)]]]),
    }
    y = {
        "k0": False,
        "k1": "v2",
        "k2": 321,
        "k3": 1.24,
    }

    objcls = Dictionary
    permutations = [
        (x, x),
        (x, y),
        (y, y),
        (y, x),
    ]

    # test supported methods
    for op in ["__eq__", "__ne__"]:
        _test_generic(objcls, permutations, op)

    # test unsupported methods
    for op in [
        "__gt__",
        "__ge__",
        "__lt__",
        "__le__",
        "__and__",
        "__or__",
        "__xor__",
        "intersects",
        "inside",
        "outside",
    ]:
        _test_unsupported(objcls, permutations, op)

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none()  # type: ignore - issue #604
    assert not v1.is_not_none()  # type: ignore - issue #604
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert not v2.is_none()  # type: ignore - issue #604
    assert v2.is_not_none()  # type: ignore - issue #604

    # test encoding
    assert {
        "k0": True,
        "k1": "v1",
        "k2": 123,
        "k3": 1.24,
        "k4": {"type": "datetime", "value": "2024-01-01T00:00:00"},
        "k5": {"type": "date", "value": "2024-01-01"},
        "k6": {"type": "time", "value": "01:00:00"},
        "k7": {"type": "duration", "value": 100.0},
        "k8": {
            "type": "geojson",
            "value": {"type": "Point", "coordinates": (1, -1)},
        },
        "k9": {
            "type": "geojson",
            "value": {"type": "MultiPoint", "coordinates": [(0, 0), (1, 1)]},
        },
        "k10": {
            "type": "geojson",
            "value": {"type": "LineString", "coordinates": [(0, 0), (1, 1)]},
        },
        "k11": {
            "type": "geojson",
            "value": {
                "type": "MultiLineString",
                "coordinates": [[(0, 0), (1, 1)]],
            },
        },
        "k12": {
            "type": "geojson",
            "value": {
                "type": "Polygon",
                "coordinates": [[(0, 0), (1, 1), (0, 1), (0, 0)]],
            },
        },
        "k13": {
            "type": "geojson",
            "value": {
                "type": "MultiPolygon",
                "coordinates": [[[(0, 0), (1, 1), (0, 1), (0, 0)]]],
            },
        },
    } == Dictionary(x).encode_value()



from valor import enums, schemas


def test_evaluation_request():
    params = {
        "dataset_names": ["ds"],
        "model_names": ["md"],
        "filters": {},
        "parameters": {
            "task_type": enums.TaskType.OBJECT_DETECTION.value,
            "convert_annotations_to_type": enums.AnnotationType.BOX.value,
        },
    }
    schemas.EvaluationRequest(**params)



from unittest.mock import patch

import pytest

from valor import Client


@patch("time.sleep")
@patch("valor.Client")
def test_timeouts(mock_sleep, mock_client):

    with pytest.raises(TimeoutError):
        Client.delete_dataset(mock_client, name="some_dataset", timeout=1)

    with pytest.raises(TimeoutError):
        Client.delete_model(mock_client, name="some_dataset", timeout=1)

    with pytest.raises(TimeoutError):
        Client.delete_evaluation(mock_client, evaluation_id=1, timeout=1)



from typing import Optional

from PIL.Image import Image

from valor import Datum
from valor.schemas import Integer


class ImageMetadata:
    """
    A class describing the metadata for a particular image.

    Attributes
    ----------
    datum : valor.Datum
        The underlying datum object.
    height: int
        The height of the image in pixels.
    width: int
        The width of the image in pixels.
    """

    def __init__(self, datum: Datum):
        """
        Creates an `ImageMetadata` object from a `valor.Datum`.

        Parameters
        ----------
        datum : Datum
            The `Datum` to extract metadata from.
        """
        if not isinstance(datum, Datum):
            raise TypeError
        elif datum.is_symbolic:
            raise ValueError

        self.datum = datum

    @classmethod
    def create(
        cls,
        uid: str,
        height: int,
        width: int,
        metadata: Optional[dict] = None,
    ):
        """
        Create an instance of ImageMetadata using components.

        Parameters
        ----------
        uid: str
            The uid of the image.
        height: int
            The height of the image in pixels.
        width: int
            The width of the image in pixels.
        metadata: dict, optional
            Any other associated metadata.
        """
        if not isinstance(height, int) or not isinstance(width, int):
            raise TypeError("Height and width must be integers.")
        metadata = metadata if metadata else dict()
        metadata["height"] = height
        metadata["width"] = width
        return cls(
            datum=Datum(
                uid=uid,
                metadata=metadata,
            )
        )

    @classmethod
    def from_pil(cls, image: Image, uid: str):
        """
        Creates an `ImageMetadata` object from an image.

        Parameters
        ----------
        image : PIL.Image.Image
            The image to create metadata for.
        uid : str
            The UID of the image.
        """
        width, height = image.size
        return cls.create(uid=uid, height=height, width=width)

    @property
    def height(self) -> int:
        """Returns image height in pixels."""
        value = self.datum.metadata["height"]
        if not isinstance(value, int):
            raise TypeError
        return int(value)

    @property
    def width(self) -> int:
        """Returns image width in pixels."""
        value = self.datum.metadata["width"]
        if not isinstance(value, int):
            raise TypeError
        return int(value)


class VideoFrameMetadata:
    """
    A class describing the metadata for the frame of a video.

    Attributes
    ----------
    datum : valor.Datum
        The underlying datum object.
    height: int
        The height of the image in pixels.
    width: int
        The width of the image in pixels.
    frame: int
        The number of seconds into the video that the frame was taken.
    """

    def __init__(self, datum: Datum):
        """
        Creates a `VideoFrameMetadata` object from a `Datum`.

        Parameters
        ----------
        datum : Datum
            The `Datum` to extract metadata from.
        """
        if not isinstance(datum, Datum):
            raise TypeError
        elif datum.is_symbolic:
            raise ValueError

        height = int(datum.metadata.get_value()["height"].get_value())
        width = int(datum.metadata.get_value()["width"].get_value())
        frame = int(datum.metadata.get_value()["frame"].get_value())
        datum.metadata["height"] = Integer(height)
        datum.metadata["width"] = Integer(width)
        datum.metadata["frame"] = Integer(frame)
        self.datum = datum

    @classmethod
    def create(
        cls,
        uid: str,
        height: int,
        width: int,
        frame: int,
        metadata: Optional[dict] = None,
    ):
        """
        Create an instance of VideoFrameMetadata using components.

        Parameters
        ----------
        uid: str
            The uid of the image.
        height: int
            The height of the image in pixels.
        width: int
            The width of the image in pixels.
        frame: int
            The video frame index.
        metadata: dict, optional
            Any other associated metadata.
        """
        if (
            not isinstance(height, int)
            or not isinstance(width, int)
            or not isinstance(frame, int)
        ):
            raise TypeError("Height, width and frame must be integers.")
        metadata = metadata if metadata else dict()
        metadata["height"] = height
        metadata["width"] = width
        metadata["frame"] = frame
        return cls(
            Datum(
                uid=uid,
                metadata=metadata,
            )
        )

    @property
    def height(self) -> int:
        """Returns image height in pixels."""
        value = self.datum.metadata["height"].get_value()
        if not isinstance(value, int):
            raise TypeError
        return int(value)

    @property
    def width(self) -> int:
        """Returns image width in pixels."""
        value = self.datum.metadata["width"].get_value()
        if not isinstance(value, int):
            raise TypeError
        return int(value)

    @property
    def frame(self) -> int:
        """Returns the video frame number."""
        value = self.datum.metadata["frame"].get_value()
        if not isinstance(value, int):
            raise TypeError
        return int(value)



import numpy as np

from valor import (
    Annotation,
    Dataset,
    Datum,
    GroundTruth,
    Label,
    Model,
    Prediction,
)
from valor.schemas import (
    Box,
    ContextList,
    Dictionary,
    Float,
    List,
    Polygon,
    Raster,
    String,
)


def test_label_typing():
    assert type(Label.key) is String
    assert type(Label.value) is String
    assert type(Label.score) is Float

    label = Label(key="k1", value="v1")
    assert type(label.key) is str
    assert type(label.value) is str
    assert label.score is None

    label = Label(key="k1", value="v1", score=None)
    assert type(label.key) is str
    assert type(label.value) is str
    assert label.score is None

    label = Label(key="k1", value="v1", score=1.0)
    assert type(label.key) is str
    assert type(label.value) is str
    assert type(label.score) is float


def test_annotation_typing():
    assert type(Annotation.labels) is List[Label]
    assert type(Annotation.metadata) is Dictionary
    assert type(Annotation.bounding_box) is Box
    assert type(Annotation.polygon) is Polygon
    assert type(Annotation.raster) is Raster
    assert type(Annotation.text) is String
    assert type(Annotation.context_list) is ContextList

    annotation = Annotation(
        labels=[],
    )
    assert type(annotation.labels) is List[Label]
    assert type(annotation.metadata) is Dictionary
    assert annotation.bounding_box is None
    assert annotation.polygon is None
    assert annotation.raster is None
    assert annotation.text is None
    assert annotation.context_list is None

    bbox = Box.from_extrema(0, 1, 0, 1)
    polygon = Polygon([bbox.boundary])
    raster = Raster.from_numpy(np.zeros((10, 10)) == 0)
    annotation = Annotation(
        labels=[],
        metadata={},
        bounding_box=bbox,
        polygon=polygon,
        raster=raster,
    )
    assert type(annotation.labels) is List[Label]
    assert type(annotation.metadata) is Dictionary
    assert type(annotation.bounding_box) is Box
    assert type(annotation.polygon) is Polygon
    assert type(annotation.raster) is Raster
    assert annotation.text is None
    assert annotation.context_list is None

    text = "Example text."
    context_list = ["context 1", "context 2"]
    annotation = Annotation(
        metadata={},
        text=text,
        context_list=context_list,
    )

    assert type(annotation.labels) is List[Label]
    assert type(annotation.metadata) is Dictionary
    assert annotation.bounding_box is None
    assert annotation.polygon is None
    assert annotation.raster is None
    assert type(annotation.text) is str
    assert type(annotation.context_list) is ContextList


def test_datum_typing():
    assert type(Datum.uid) is String
    assert type(Datum.metadata) is Dictionary
    assert type(Datum.text) is String

    datum = Datum(uid="test")
    assert type(datum.uid) is str
    assert type(datum.metadata) is Dictionary
    assert datum.text is None

    text = "Example text."
    datum = Datum(uid="test", text=text, metadata={})
    assert type(datum.uid) is str
    assert type(datum.metadata) is Dictionary
    assert type(datum.text) is str


def test_groundtruth_typing():
    # GroundTruth doesn't use special properties.
    groundtruth = GroundTruth(datum=Datum(uid="uid"), annotations=[])
    assert type(groundtruth.datum) is Datum
    assert type(groundtruth.annotations) is List[Annotation]


def test_prediction_typing():
    # Prediction doesn't use special properties.
    prediction = Prediction(datum=Datum(uid="uid"), annotations=[])
    assert type(prediction.datum) is Datum
    assert type(prediction.annotations) is List[Annotation]


def test_dataset_typing():
    assert type(Dataset.name) is String
    assert type(Dataset.metadata) is Dictionary

    dataset = Dataset(name="test")
    assert type(dataset.name) is str
    assert type(dataset.metadata) is Dictionary

    dataset = Dataset(name="test", metadata={})
    assert type(dataset.name) is str
    assert type(dataset.metadata) is Dictionary


def test_model_typing():
    assert type(Model.name) is String
    assert type(Model.metadata) is Dictionary

    model = Model(name="test")
    assert type(model.name) is str
    assert type(model.metadata) is Dictionary

    model = Model(name="test", metadata={})
    assert type(model.name) is str
    assert type(model.metadata) is Dictionary



from typing import List, Tuple

import numpy as np
import pytest

from valor import Label
from valor.schemas import Box, Polygon, Raster


@pytest.fixture
def labels() -> List[Label]:
    return [
        Label(key="k1", value="v1"),
        Label(key="k2", value="v2"),
    ]


@pytest.fixture
def box_points() -> List[Tuple[float, float]]:
    return [
        (0, 0),
        (10, 0),
        (10, 10),
        (0, 10),
        (0, 0),
    ]


@pytest.fixture
def bbox() -> Box:
    return Box.from_extrema(xmin=0, xmax=10, ymin=0, ymax=10)


@pytest.fixture
def polygon(box_points) -> Polygon:
    return Polygon([box_points])


@pytest.fixture
def raster_raw_mask() -> np.ndarray:
    """
    Creates a 2d numpy of bools of shape:
    | T  F |
    | F  T |
    """
    ones = np.ones((10, 10))
    zeros = np.zeros((10, 10))
    top = np.concatenate((ones, zeros), axis=1)
    bottom = np.concatenate((zeros, ones), axis=1)
    return np.concatenate((top, bottom), axis=0) == 1


@pytest.fixture
def raster(raster_raw_mask) -> Raster:
    return Raster.from_numpy(raster_raw_mask)


@pytest.fixture
def metadata() -> dict:
    return {
        "a": 1234,
        "b": 1.234,
        "c": "1234",
    }



from enum import Enum
from typing import Set


class AnnotationType(str, Enum):
    NONE = "none"
    BOX = "box"
    POLYGON = "polygon"
    MULTIPOLYGON = "multipolygon"
    RASTER = "raster"


class TaskType(str, Enum):
    SKIP = "skip"
    EMPTY = "empty"
    CLASSIFICATION = "classification"
    OBJECT_DETECTION = "object-detection"
    SEMANTIC_SEGMENTATION = "semantic-segmentation"
    EMBEDDING = "embedding"
    TEXT_GENERATION = "text-generation"


class TableStatus(str, Enum):
    CREATING = "creating"
    FINALIZED = "finalized"
    DELETING = "deleting"


class EvaluationStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    DONE = "done"
    FAILED = "failed"
    DELETING = "deleting"


class MetricType(str, Enum):

    Accuracy = ("Accuracy",)
    Precision = ("Precision",)
    Recall = ("Recall",)
    F1 = ("F1",)
    ROCAUC = ("ROCAUC",)
    AP = "AP"
    AR = "AR"
    mAP = "mAP"
    mAR = "mAR"
    APAveragedOverIOUs = "APAveragedOverIOUs"
    mAPAveragedOverIOUs = "mAPAveragedOverIOUs"
    IOU = "IOU"
    mIOU = "mIOU"
    PrecisionRecallCurve = "PrecisionRecallCurve"
    DetailedPrecisionRecallCurve = "DetailedPrecisionRecallCurve"
    AnswerCorrectness = "AnswerCorrectness"
    AnswerRelevance = "AnswerRelevance"
    Bias = "Bias"
    BLEU = "BLEU"
    ContextPrecision = "ContextPrecision"
    ContextRecall = "ContextRecall"
    ContextRelevance = "ContextRelevance"
    Faithfulness = "Faithfulness"
    Hallucination = "Hallucination"
    ROUGE = "ROUGE"
    SummaryCoherence = "SummaryCoherence"
    Toxicity = "Toxicity"

    @classmethod
    def classification(cls) -> Set["MetricType"]:
        """
        MetricTypes for classification tasks.
        """
        return {
            cls.Accuracy,
            cls.Precision,
            cls.Recall,
            cls.F1,
            cls.ROCAUC,
            cls.PrecisionRecallCurve,
            cls.DetailedPrecisionRecallCurve,
        }

    @classmethod
    def object_detection(cls) -> Set["MetricType"]:
        """
        MetricTypes for object-detection tasks.
        """
        return {
            cls.AP,
            cls.AR,
            cls.mAP,
            cls.mAR,
            cls.APAveragedOverIOUs,
            cls.mAPAveragedOverIOUs,
            cls.PrecisionRecallCurve,
            cls.DetailedPrecisionRecallCurve,
        }

    @classmethod
    def semantic_segmentation(cls) -> Set["MetricType"]:
        """
        MetricTypes for semantic-segmentation tasks.
        """
        return {
            cls.IOU,
            cls.mIOU,
            cls.Precision,
            cls.Recall,
            cls.F1,
        }

    @classmethod
    def text_generation(cls) -> Set["MetricType"]:
        """
        MetricTypes for text-generation tasks.
        """
        return {
            cls.AnswerCorrectness,
            cls.AnswerRelevance,
            cls.Bias,
            cls.BLEU,
            cls.ContextPrecision,
            cls.ContextRecall,
            cls.ContextRelevance,
            cls.Faithfulness,
            cls.Hallucination,
            cls.ROUGE,
            cls.SummaryCoherence,
            cls.Toxicity,
        }


class ROUGEType(str, Enum):
    ROUGE1 = "rouge1"
    ROUGE2 = "rouge2"
    ROUGEL = "rougeL"
    ROUGELSUM = "rougeLsum"



import warnings
from typing import Any, Optional, Union


class Condition:
    """Base class for defining a conditional operation."""

    def __init__(self, lhs: Any, rhs: Optional[Any] = None) -> None:
        """
        Create a condition.

        Parameters
        ----------
        lhs : Variable
            A variable.
        rhs : Variable, optional
            An optional rhs variable.
        """
        # validate lhs
        if not lhs.is_symbolic:
            warnings.warn(
                "Values are currently not supported as the lhs operand in the API.",
                RuntimeWarning,
            )

        # validate rhs - symbols are not current supported
        if rhs is not None:
            if rhs.is_symbolic:
                warnings.warn(
                    "Symbols are currently not supported as the rhs operand in the API.",
                    RuntimeWarning,
                )

        self.lhs = lhs
        self.rhs = rhs
        self.op = type(self).__name__.lower()

    def __and__(self, other: Any):
        return And(self, other)

    def __or__(self, other: Any):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def to_dict(self):
        return {
            "lhs": self.lhs.to_dict(),
            "rhs": self.rhs.to_dict() if self.rhs is not None else None,
            "op": self.op,
        }


class Function:
    """Base class for defining a logical function."""

    def __init__(self, *args) -> None:
        if len(args) == 0:
            raise ValueError("Expected at least one argument.")

        self._args = []
        for arg in args:
            if not hasattr(arg, "to_dict"):
                raise ValueError(
                    f"Arguments should be symbolic or functional. Received '{arg}'."
                )
            if isinstance(arg, type(self)):
                self._args.extend(arg._args)
            else:
                self._args.append(arg)
        self._args = self._args if len(self._args) > 1 else self._args[0]

    def __repr__(self):
        args = ", ".join([arg.__repr__() for arg in self._args])
        return f"{type(self).__name__}({args})"

    def __str__(self):
        values = [arg.__repr__() for arg in self._args]
        args = ", ".join(values)
        return f"{type(self).__name__}({args})"

    def __and__(self, other: Any):
        return And(self, other)

    def __or__(self, other: Any):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def to_dict(self):
        """Encode to a JSON-compatible dictionary."""
        args = (
            [arg.to_dict() for arg in self._args]
            if isinstance(self._args, list)
            else self._args.to_dict()
        )
        return {"op": type(self).__name__.lower(), "args": args}


class And(Function):
    """
    Implementation of logical AND (&).

    This class represents a logical AND operation that can be performed on
    two or more arguments. It supports chaining of multiple AND operations
    using the `&` operator.

    Parameters
    ----------
    *args : Any
        The arguments to be logically ANDed together. At least two arguments
        are required.

    Raises
    ------
    ValueError
        If fewer than two arguments are provided.

    Examples
    --------
    >>> a = And(Label.key == "k1", Label.value == "v1")
    >>> b = And(Label.key == "k1", Label.value == "v2")
    >>> c = a & b

    Methods
    -------
    __and__(other)
        Supports chaining of multiple `And` operations using the `&` operator.
    """

    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError("Expected at least two arguments.")
        super().__init__(*args)

    def __and__(self, other: Any):
        if isinstance(other, And):
            self._args.extend(other._args)
        else:
            self._args.append(other)
        return self


class Or(Function):
    """Implementation of logical OR (|).

    This class represents a logical OR operation that can be performed on
    two or more arguments. It supports chaining of multiple OR operations
    using the `|` operator.

    Parameters
    ----------
    *args : Any
        The arguments to be logically ORed together. At least two arguments
        are required.

    Raises
    ------
    ValueError
        If fewer than two arguments are provided.

    Examples
    --------
    >>> a = Or(Label.key == "k1", Label.key == "k2")
    >>> b = Or(Label.value == "v1", Label.value == "v2")
    >>> c = a | b

    Methods
    -------
    __or__(other)
        Supports chaining of multiple `Or` operations using the `|` operator.
    """

    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError("Expected at least two arguments.")
        super().__init__(*args)

    def __or__(self, other: Any):
        if isinstance(other, Or):
            self._args.extend(other._args)
        else:
            self._args.append(other)
        return self


class Not(Function):
    """Implementation of logical negation (~).

    This class represents a logical NOT operation that can be performed on
    on a single arguments. It supports chaining of multiple NOT operations
    using the `~` operator.

    Parameters
    ----------
    *args : Any
        The arguments to be logically ORed together. At least two arguments
        are required.

    Raises
    ------
    ValueError
        If the number of args is not equal to one.

    Examples
    --------
    >>> a = Not(Label.key == "k1")
    >>> b = ~a

    Methods
    -------
    __invert__()
        Supports chaining of multiple `Not` operations using the `~` operator.
    """

    def __init__(self, *args):
        if len(args) != 1:
            raise ValueError("Negation only takes one argument.")
        elif isinstance(args[0], Not):
            return args[0]._args
        super().__init__(*args)

    def __invert__(self):
        """Inverts negation so return contents."""
        if isinstance(self._args, list):
            raise ValueError("Negation only takes one argument.")
        return self._args


class IsNull(Condition):
    """Implementation of is null value check."""

    pass


class IsNotNull(Condition):
    """Implementation of is not null value check."""

    pass


class Eq(Condition):
    """Implementation of the equality operator '=='."""

    pass


class Ne(Condition):
    """Implementation of the inequality operator '!='."""

    def to_dict(self):
        return Not(Eq(lhs=self.lhs, rhs=self.rhs)).to_dict()


class Gt(Condition):
    """Implementation of the greater-than operator '>'."""

    pass


class Gte(Condition):
    """Implementation of the greater-than or equal operator '>='."""

    pass


class Lt(Condition):
    """Implementation of the less-than operator '<'."""

    pass


class Lte(Condition):
    """Implementation of the less-than or equal operator '<='."""

    pass


class Intersects(Condition):
    """Implementation of the spatial 'intersects' operator."""

    pass


class Inside(Condition):
    """Implementation of the spatial 'inside' operator."""

    pass


class Outside(Condition):
    """Implementation of the spatial 'outside' operator."""

    pass


class Contains(Condition):
    """Implementation of the list 'contains' operator."""

    pass


FunctionType = Union[
    And,
    Or,
    Not,
    IsNull,
    IsNotNull,
    Eq,
    Ne,
    Gt,
    Gte,
    Lt,
    Lte,
    Intersects,
    Inside,
    Outside,
    Contains,
]



import datetime

from valor import Annotation, Dataset, Datum, Filter, Label, Model
from valor.schemas import And


def test_empty_filter():
    assert Filter().to_dict() == {
        "datasets": None,
        "models": None,
        "datums": None,
        "annotations": None,
        "groundtruths": None,
        "predictions": None,
        "labels": None,
        "embeddings": None,
    }


def test_declarative_filtering():
    filters = Filter(
        datums=And(
            Datum.uid == "uid1",
            Datum.metadata["arbitrary_datetime_key"]
            >= datetime.timedelta(days=1),
            Datum.metadata["arbitrary_datetime_key"]
            <= datetime.timedelta(days=2),
        ),
        annotations=And(
            Dataset.name == "dataset1",
            Dataset.metadata["arbitrary_numeric_key"] >= 10,
            Dataset.metadata["arbitrary_numeric_key"] < 20,
            # geometry filters
            Annotation.raster.is_none(),
            Annotation.polygon.is_none(),
            Annotation.bounding_box.is_not_none(),
            Annotation.bounding_box.area >= 1000,
            Annotation.bounding_box.area <= 5000,
            Annotation.metadata["myKey"] == "helloworld",
            # label filters
            Label.key == "k2",
            Label.value == "v2",
        ),
        labels=And(
            Label.key == "k1",
            Label.score > 0.5,
            Label.score < 0.75,
        ),
        predictions=And(
            Model.name == "model1",
            Model.metadata["arbitrary_str_key"] == "arbitrary value",
        ),
    )

    assert filters.to_dict() == {
        "datasets": None,
        "models": None,
        "datums": {
            "op": "and",
            "args": [
                {
                    "lhs": {"name": "datum.uid", "key": None},
                    "rhs": {"type": "string", "value": "uid1"},
                    "op": "eq",
                },
                {
                    "lhs": {
                        "name": "datum.metadata",
                        "key": "arbitrary_datetime_key",
                    },
                    "rhs": {"type": "duration", "value": 86400.0},
                    "op": "gte",
                },
                {
                    "lhs": {
                        "name": "datum.metadata",
                        "key": "arbitrary_datetime_key",
                    },
                    "rhs": {"type": "duration", "value": 172800.0},
                    "op": "lte",
                },
            ],
        },
        "annotations": {
            "op": "and",
            "args": [
                {
                    "lhs": {"name": "dataset.name", "key": None},
                    "rhs": {"type": "string", "value": "dataset1"},
                    "op": "eq",
                },
                {
                    "lhs": {
                        "name": "dataset.metadata",
                        "key": "arbitrary_numeric_key",
                    },
                    "rhs": {"type": "integer", "value": 10},
                    "op": "gte",
                },
                {
                    "lhs": {
                        "name": "dataset.metadata",
                        "key": "arbitrary_numeric_key",
                    },
                    "rhs": {"type": "integer", "value": 20},
                    "op": "lt",
                },
                {
                    "lhs": {"name": "annotation.raster", "key": None},
                    "rhs": None,
                    "op": "isnull",
                },
                {
                    "lhs": {"name": "annotation.polygon", "key": None},
                    "rhs": None,
                    "op": "isnull",
                },
                {
                    "lhs": {"name": "annotation.bounding_box", "key": None},
                    "rhs": None,
                    "op": "isnotnull",
                },
                {
                    "lhs": {
                        "name": "annotation.bounding_box.area",
                        "key": None,
                    },
                    "rhs": {"type": "float", "value": 1000},
                    "op": "gte",
                },
                {
                    "lhs": {
                        "name": "annotation.bounding_box.area",
                        "key": None,
                    },
                    "rhs": {"type": "float", "value": 5000},
                    "op": "lte",
                },
                {
                    "lhs": {"name": "annotation.metadata", "key": "mykey"},
                    "rhs": {"type": "string", "value": "helloworld"},
                    "op": "eq",
                },
                {
                    "lhs": {"name": "label.key", "key": None},
                    "rhs": {"type": "string", "value": "k2"},
                    "op": "eq",
                },
                {
                    "lhs": {"name": "label.value", "key": None},
                    "rhs": {"type": "string", "value": "v2"},
                    "op": "eq",
                },
            ],
        },
        "groundtruths": None,
        "predictions": {
            "op": "and",
            "args": [
                {
                    "lhs": {"name": "model.name", "key": None},
                    "rhs": {"type": "string", "value": "model1"},
                    "op": "eq",
                },
                {
                    "lhs": {
                        "name": "model.metadata",
                        "key": "arbitrary_str_key",
                    },
                    "rhs": {"type": "string", "value": "arbitrary value"},
                    "op": "eq",
                },
            ],
        },
        "labels": {
            "op": "and",
            "args": [
                {
                    "lhs": {"name": "label.key", "key": None},
                    "rhs": {"type": "string", "value": "k1"},
                    "op": "eq",
                },
                {
                    "lhs": {"name": "label.score", "key": None},
                    "rhs": {"type": "float", "value": 0.5},
                    "op": "gt",
                },
                {
                    "lhs": {"name": "label.score", "key": None},
                    "rhs": {"type": "float", "value": 0.75},
                    "op": "lt",
                },
            ],
        },
        "embeddings": None,
    }
